# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import threading
from typing import Optional

import importlib_metadata
import numpy as np
import qtawesome as qta

from PySide6 import QtCore
from PySide6.QtGui import QPainter, QPaintEvent, QTextDocument
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QStatusBar,
    QStyle,
    QStyleOption,
    QTextBrowser,
    QWidget,
)

from acconeer.exptool.app.new import check_package_outdated, get_latest_changelog
from acconeer.exptool.app.new.app_model import AppModel

from .misc import BUTTON_ICON_COLOR


class FrameCountLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Number of received frames")

        app_model.sig_frame_count.connect(self._on_app_model_frame_count)
        self._on_app_model_frame_count(None)

    def _on_app_model_frame_count(self, frame_count: Optional[int]) -> None:
        if frame_count is None:
            css = "color: #888;"
            text = "Frames: -    "
        else:
            css = ""
            text = f"Frames: {frame_count:5}"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)


class RateStatsLabel(QLabel):
    _FPS: int = 10

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Reported update rate")

        self.rate = np.nan
        self.rate_warning = False

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_rate_stats.connect(self._on_app_model_rate_stats)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if np.isnan(self.rate):
            css = "color: #888;"
            text = "     - Hz"
        else:
            css = "color: #FD5200;" if self.rate_warning else ""
            text = f"{self.rate:>6.1f} Hz"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)

    def _on_app_model_rate_stats(
        self,
        rate: float,
        rate_warning: bool,
        jitter: float,
        jitter_warning: bool,
    ) -> None:
        self.rate = rate
        self.rate_warning = rate_warning


class JitterStatsLabel(QLabel):
    _FPS: int = 10

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Reported jitter")

        self.jitter = np.nan
        self.jitter_warning = False

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_rate_stats.connect(self._on_app_model_rate_stats)

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if np.isnan(self.jitter):
            css = "color: #888;"
            text = "    - ms"
        else:
            css = "color: #FD5200;" if self.jitter_warning else ""
            text = f"{self.jitter * 1e3:5.1f} ms"

        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)

    def _on_app_model_rate_stats(
        self,
        rate: float,
        rate_warning: bool,
        jitter: float,
        jitter_warning: bool,
    ) -> None:
        self.jitter = jitter
        self.jitter_warning = jitter_warning


class BackendCPUPercentLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setToolTip("Client process CPU load")

        app_model.sig_backend_cpu_percent.connect(self._on_app_model_backend_cpu_percent)

        self._on_app_model_backend_cpu_percent(0)

    def _on_app_model_backend_cpu_percent(self, cpu_percent: int) -> None:
        self.setText(f"CPU: {cpu_percent:3}%")

        css = "color: #FD5200;" if cpu_percent >= 85 else ""
        self.setStyleSheet(f"QWidget{{{css}}}")


class RSSVersionLabel(QLabel):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        app_model.sig_notify.connect(self._on_app_model_update)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if app_model.rss_version is None:
            css = "color: #888;"
            text = "RSS: <not connected>"
            tooltip = ""
        else:
            text = f"RSS: {app_model.rss_version}"

            if app_model.connection_warning:
                css = "background-color: #D78100; color: #e2e2e2;"
                tooltip = app_model.connection_warning
            else:
                css = ""
                tooltip = ""

        self.setToolTip(tooltip)
        self.setStyleSheet(f"QWidget{{{css}}}")
        self.setText(text)


class VersionLabel(QWidget):
    sig_version_outdated = QtCore.Signal()

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.parent = parent
        h_layout = QHBoxLayout()
        self.setLayout(h_layout)
        h_layout.setContentsMargins(0, 0, 0, 0)
        h_layout.setSpacing(0)

        et_version = importlib_metadata.version("acconeer-exptool")
        et_version_text = f"ET: {et_version}"
        et_version_label = QLabel(et_version_text, self)
        h_layout.addWidget(et_version_label)

        def version_check(et_version: str) -> None:
            version_outdated, latest_v = check_package_outdated("acconeer-exptool", et_version)
            self.changelog = None

            if version_outdated:
                self.changelog = get_latest_changelog()
                self.setStyleSheet(
                    f"QWidget{{background-color: {BUTTON_ICON_COLOR}; color: #e2e2e2}}"
                )

                self.setToolTip(
                    "There is a new software version available!\n"
                    f"The latest version is: {latest_v}. \n"
                    "Click to view changelog."
                )
                self.sig_version_outdated.emit()
                self.mousePressEvent = self._toggle_changelog

        self.sig_version_outdated.connect(self._create_changelog_window)
        self.sig_version_outdated.connect(lambda: self._add_icon_to_version_label(self))

        version_thread = threading.Thread(target=lambda: version_check(et_version))
        version_thread.start()

    def _create_changelog_window(self) -> None:
        if self.changelog is not None:
            self.cl_window = ChangelogWindow(self)
            self.cl_window.set_text(self.changelog)

    def _add_icon_to_version_label(self, version_widget: QWidget) -> None:
        layout = version_widget.layout()
        icon_widget = qta.IconWidget()
        icon_widget.setIcon(qta.icon("fa.refresh", color="#e2e2e2"))
        layout.addWidget(icon_widget)

    def _toggle_changelog(self, label: QLabel) -> None:
        if self.changelog is not None:
            self.cl_window.setVisible(not self.cl_window.isVisible())

    def paintEvent(self, pe: QPaintEvent) -> None:
        o = QStyleOption()
        o.initFrom(self)
        p = QPainter(self)
        self.style().drawPrimitive(QStyle.PE_Widget, o, p, self)


class StatusBar(QStatusBar):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.parent = parent

        app_model.sig_status_message.connect(self._on_app_model_status_message)

        self.message_timer = QtCore.QTimer(self)
        self.message_timer.setInterval(3000)
        self.message_timer.setSingleShot(True)
        self.message_timer.timeout.connect(self._on_timer)

        self.message_widget = QLabel(self)
        self.addWidget(self.message_widget)

        self.addPermanentWidget(FrameCountLabel(app_model, self))
        self.addPermanentWidget(RateStatsLabel(app_model, self))
        self.addPermanentWidget(JitterStatsLabel(app_model, self))
        self.addPermanentWidget(BackendCPUPercentLabel(app_model, self))
        self.addPermanentWidget(RSSVersionLabel(app_model, self))
        self.addPermanentWidget(VersionLabel(app_model, self))

        font_families = [
            "Consolas",  # Windows
            "Droid Sans Mono",  # Ubuntu
            "DejaVu Sans Mono",  # Ubuntu, backup
            "SF Mono",  # Mac
        ]
        font_family = ", ".join(f'"{ff}"' for ff in font_families)
        self.setStyleSheet(f"QWidget{{font-family: {font_family};}}")

    def _on_app_model_status_message(self, message: Optional[str]) -> None:
        self.message_timer.stop()

        if message:
            self.message_widget.setText(message)
            self.message_timer.start()
        else:
            self.message_widget.clear()

    def _on_timer(self) -> None:
        self.message_widget.clear()


class ChangelogWindow(QMainWindow):
    def __init__(self, parent: QWidget = None) -> None:
        super().__init__()

        self.setWindowTitle("Changelog")
        self.text_browser = QTextBrowser()
        self.setCentralWidget(self.text_browser)
        self.set_center(parent.parent.parent)
        self.showEvent = lambda _: self.set_center(parent.parent.parent)

    def set_text(self, text: str) -> None:
        document = QTextDocument()
        document.setMarkdown(text)
        self.text_browser.setDocument(document)

    def set_center(self, main_window: QMainWindow) -> None:
        fg = self.frameGeometry()
        fg.moveCenter(main_window.geometry().center())
        self.move(fg.topLeft())
        self.resize(500, 400)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.IQServiceConfig()
    config.range_interval = [0.2, 0.8]
    config.update_rate = 30
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 2

    history_length = et.configbase.IntParameter(
        default_value=100,
        limits=(10, 1000),
        label="History length",
        order=0,
    )

    sf = et.configbase.FloatParameter(
        label="Smoothing factor",
        default_value=None,
        limits=(0.1, 0.999),
        decimals=3,
        optional=True,
        optional_label="Enable filter",
        optional_default_set_value=0.9,
        updateable=True,
        order=10,
    )


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        depths = et.a111.get_range_depths(sensor_config, session_info)
        num_depths = depths.size
        num_sensors = len(sensor_config.sensor)
        history_length = processing_config.history_length
        self.history = np.zeros([history_length, num_sensors, num_depths], dtype="complex")
        self.lp_data = np.zeros([num_sensors, num_depths], dtype="complex")
        self.update_index = 0
        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.sf = processing_config.sf if processing_config.sf is not None else 0.0

    def dynamic_sf(self, static_sf):
        return min(static_sf, 1.0 - 1.0 / (1.0 + self.update_index))

    def process(self, data, data_info):
        self.history = np.roll(self.history, -1, axis=0)
        self.history[-1] = data

        sf = self.dynamic_sf(self.sf)
        self.lp_data = sf * self.lp_data + (1 - sf) * data

        self.update_index += 1

        return {
            "data": self.lp_data,
            "history": self.history,
        }



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.SparseServiceConfig()
    config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_3
    config.range_interval = [0.48, 0.72]
    config.sweeps_per_frame = 16
    config.hw_accelerated_average_samples = 60
    config.update_rate = 60
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 1
    WINDOW_SIZE_POW_OF_2_MAX = 12
    ROLLING_HISTORY_SIZE_MAX = 1000

    show_time_domain = et.configbase.BoolParameter(
        label="Show data in time domain",
        default_value=True,
        updateable=True,
        order=0,
    )

    show_spect_history = et.configbase.BoolParameter(
        label="Show spectrum history",
        default_value=False,
        updateable=True,
        order=10,
    )

    show_depthwise_spect = et.configbase.BoolParameter(
        label="Show depthwise spectrum",
        default_value=False,
        updateable=True,
        order=20,
    )

    window_size_pow_of_2 = et.configbase.FloatParameter(
        label="Window size, power of 2",
        default_value=8,
        limits=(3, WINDOW_SIZE_POW_OF_2_MAX),
        decimals=0,
        updateable=True,
        order=100,
    )

    _window_size = et.configbase.get_virtual_parameter_class(et.configbase.IntParameter)(
        label="Window size",
        get_fun=lambda conf: 2 ** int(conf.window_size_pow_of_2),
        visible=False,
    )

    overlap = et.configbase.FloatParameter(
        label="Overlap",
        default_value=0.95,
        limits=(0, 1),
        updateable=True,
        order=200,
    )

    rolling_history_size = et.configbase.FloatParameter(
        label="Rolling history size",
        default_value=100,
        decimals=0,
        logscale=True,
        limits=(10, ROLLING_HISTORY_SIZE_MAX),
        updateable=True,
        order=300,
    )

    def check(self):
        alerts = super().check()

        msg = "{}".format(self._window_size)
        alerts.append(et.configbase.Info("window_size_pow_of_2", msg))

        return alerts


class Processor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.f = sensor_config.update_rate
        depths = et.a111.get_range_depths(sensor_config, session_info)
        self.num_depths = depths.size

        max_window_size = 2**ProcessingConfiguration.WINDOW_SIZE_POW_OF_2_MAX
        self.sweep_history = np.full([max_window_size, self.num_depths], np.nan)

        self.collapsed_asd = None
        self.collapsed_asd_history = None

        self.window_size = None
        self.frames_between_updates = None
        self.rolling_history_size = None

        self.tick_idx = 0
        self.last_update_tick = 0

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        invalid = self.window_size != processing_config._window_size

        self.window_size = processing_config._window_size
        self.frames_between_updates = int(
            round(self.window_size * (1 - processing_config.overlap))
        )

        self.rolling_history_size = int(processing_config.rolling_history_size)

        if invalid:
            self.collapsed_asd_history = np.zeros(
                [
                    ProcessingConfiguration.ROLLING_HISTORY_SIZE_MAX,
                    self.window_size // 2,
                ]
            )

        if invalid and self.tick_idx > 0:
            self.update_spect()

    def process(self, data, data_info):
        frame = data

        mean_sweep = frame.mean(axis=0)

        self.sweep_history = np.roll(self.sweep_history, -1, axis=0)
        self.sweep_history[-1] = mean_sweep

        outdated = (self.tick_idx - self.last_update_tick) > self.frames_between_updates
        if self.tick_idx == 0 or outdated:
            self.update_spect()

        self.tick_idx += 1

        return self.gather_result()

    def update_spect(self):
        x = self.sweep_history[-self.window_size :]
        x = x - np.nanmean(x, axis=0, keepdims=True)
        x = np.nan_to_num(x)
        fft = np.fft.rfft(x.T * np.hanning(x.shape[0]), axis=1)
        asd = np.abs(fft)[:, 1:]

        self.collapsed_asd = asd.sum(axis=0)
        self.dw_asd = asd

        self.collapsed_asd_history = np.roll(self.collapsed_asd_history, -1, axis=0)
        self.collapsed_asd_history[-1] = self.collapsed_asd

        self.last_update_tick = self.tick_idx

    def gather_result(self):
        ts = np.arange(-self.window_size, 0, dtype="float") + 1
        fs = np.arange(self.window_size // 2, dtype="float") + 1

        if self.f:
            ts *= 1 / self.f
            fs *= 0.5 * self.f / fs[-1]

        cropped_history = self.collapsed_asd_history[-self.rolling_history_size :]

        return {
            "ts": ts,
            "sweep_history": self.sweep_history[-self.window_size :],
            "fs": fs,
            "collapsed_asd": self.collapsed_asd,
            "collapsed_asd_history": cropped_history,
            "dw_asd": self.dw_asd,
        }



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import argparse
import getpass
import logging
import tempfile
import time

import acconeer.exptool as et
from acconeer.exptool.flash._bin_fetcher import (
    BIN_FETCH_PROMPT,
    clear_cookies,
    download,
    get_content,
    get_cookies,
    is_redirected_to_login_page,
    login,
    save_cookies,
)
from acconeer.exptool.flash._products import (
    EVK_TO_PRODUCT_MAP,
    PRODUCT_NAME_TO_FLASH_MAP,
    PRODUCT_PID_TO_FLASH_MAP,
)

from ._dev_license import DevLicense
from ._dev_license_tui import DevLicenseTuiDialog


log = logging.getLogger(__name__)


def _query_yes_no(question: str, default: str = "yes") -> bool:
    """Ask a yes/no question via raw_input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
            It must be "yes" (the default), "no" or None (meaning
            an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        print(question + prompt, end="", flush=True)
        choice = input().lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' " "(or 'y' or 'n').\n")


def flasher_progress_callback(progress, end=False):
    print(f"\rProgress: [{progress}/100]", end="", flush=True)
    if end:
        print()


def flash_image(image_path, flash_device, device_name=None, progress_callback=None):
    if flash_device:
        serial_device_name = device_name or flash_device.name
        if serial_device_name is None:
            raise ValueError("Unknown device type")
        serial_device_name = serial_device_name.upper()
        if (
            isinstance(flash_device, et.utils.USBDevice)
            and flash_device.pid in PRODUCT_PID_TO_FLASH_MAP.keys()
        ):
            PRODUCT_PID_TO_FLASH_MAP[flash_device.pid].flash(
                flash_device, serial_device_name, image_path, progress_callback
            )
        elif (
            isinstance(flash_device, et.utils.SerialDevice)
            and serial_device_name in PRODUCT_NAME_TO_FLASH_MAP.keys()
        ):
            PRODUCT_NAME_TO_FLASH_MAP[serial_device_name].flash(
                flash_device, serial_device_name, image_path, progress_callback
            )
        else:
            raise NotImplementedError(f"No flash support device {str(flash_device)}")
    else:
        raise ValueError("No flash device")


def find_flash_device(
    device_name=None, port=None, serial_number=None, use_usb=True, use_serial=True, verbose=True
):
    all_devices = []
    found_devices = []
    flash_device = None

    if use_serial:
        all_devices.extend(et.utils.get_serial_devices())

    if use_usb:
        all_devices.extend(et.utils.get_usb_devices())

    if port is not None:
        for device in all_devices:
            if isinstance(device, et.utils.SerialDevice):
                if port == device.port:
                    return device

    for device in all_devices:
        if device_name is not None and device.name != device_name:
            # Device name did not match
            continue
        elif serial_number is not None and device.serial != serial_number:
            # Serial number did not match
            continue
        else:
            found_devices.append(device)

    if len(found_devices) == 0:
        if verbose:
            print("No devices connected")
    elif len(found_devices) > 1:
        if verbose:
            print("Found multiple Acconeer products:")
            print("".join([f" - {dev}\n" for dev in found_devices]))
    else:
        flash_device = found_devices[0]

    if flash_device is None:
        raise Exception(
            "Device couldn't be autodetected\n"
            "Specify the device by using the"
            " --port, --device, --interface or --serial-number flags."
        )

    return flash_device


def get_flash_device_from_args(args, verbose=False):
    use_usb = args.interface is None or args.interface == "usb"
    use_serial = args.interface is None or args.interface == "serial"
    flash_device = find_flash_device(
        port=args.port,
        device_name=args.device,
        serial_number=args.serial_number,
        use_usb=use_usb,
        use_serial=use_serial,
        verbose=verbose,
    )
    return flash_device


def get_flash_known_devices():
    known_devices = []
    for _, item in EVK_TO_PRODUCT_MAP.items():
        if item not in known_devices:
            known_devices.append(item)
    return known_devices


def get_flash_download_name(device, device_name):
    name = device_name or device.name
    if name in EVK_TO_PRODUCT_MAP.keys():
        return EVK_TO_PRODUCT_MAP[name]
    raise ValueError(f"Unknown device {name}")


def fetch_and_flash(args):
    try:
        flash_device = get_flash_device_from_args(args)

        cookies = get_cookies()
        new_login = False
        login_succeed = False
        while login_succeed is False:
            if cookies is None:
                print("\n" + BIN_FETCH_PROMPT)
                email = input("\nEmail: ")
                password = getpass.getpass("Password: ")
                new_login = True
                cookies = login(email, password)

            print("\nLogging in... ", end="", flush=True)
            login_succeed, session, page = get_content(cookies)
            if not login_succeed:
                cookies = None
                print("[Error]")

                if is_redirected_to_login_page(page):
                    print(
                        "Request failed. Try downloading the image directly "
                        "from https://developer.acconeer.com"
                    )

            else:
                print("[OK]")

        if new_login:
            if _query_yes_no(
                "\nWe use cookies to optimize the service of our applications."
                "\n\nCookie Policy: https://acconeer.com/cookie-policy-eu/"
                "\nPrivacy Statement: https://developer.acconeer.com/privacy-policy/"
                "\n\nSelecting yes to the following question will store a cookie on "
                "your computer:"
                "\nRemember me?"
            ):
                save_cookies(cookies)

        with tempfile.TemporaryDirectory() as tmp_dir:
            print("Preparing license agreement... ", end="", flush=True)
            license = DevLicense()
            license.load()
            print("[OK]")

            time.sleep(1)
            DevLicenseTuiDialog.run(log="textual.log", license=license)
            license_accepted = DevLicenseTuiDialog.get_accept()

            if license_accepted:
                device_name = get_flash_download_name(flash_device, args.device)
                try:
                    print(f"Downloading {device_name} image file... ", end="", flush=True)
                    bin_path, version = download(
                        session=session,
                        cookies=cookies,
                        path=tmp_dir,
                        device=device_name,
                    )
                    print(f"[OK] (version: {version})")

                    try:
                        if _query_yes_no("Proceed to flashing you device?"):
                            print("Flashing...")
                            flash_image(
                                bin_path,
                                flash_device,
                                device_name=device_name,
                                progress_callback=flasher_progress_callback,
                            )
                    except ValueError:
                        print("No devices found, try connecting a device before flashing.")
                except Exception as e:
                    print("[Error]")
                    log.error(str(e))
            else:
                print("You need to accept the license agreement to download the image file.")

        session.close()

    except Exception as e:
        print(str(e))


def main():
    parser = argparse.ArgumentParser(description="Image Flasher")
    subparsers = parser.add_subparsers(help="sub-command help", dest="operation", required=True)
    subparsers.add_parser("list", help="List connected devices")
    subparser = subparsers.add_parser("flash", help="Flash device")

    subparser.add_argument(
        "--port",
        dest="port",
        help="Serial port. Only used if device type can't be autodetected.",
        type=str,
    )
    subparser.add_argument(
        "--device",
        "-d",
        dest="device",
        help="Device type. Only used if device type can't be autodetected.",
        type=str.upper,
        choices=["XC120", "XE125", "XM125"],
    )
    subparser.add_argument(
        "--serial-number",
        "-sn",
        dest="serial_number",
        help="Device serial number. Only used if device type can't be autodetected.",
        type=str,
    )
    subparser.add_argument(
        "--interface",
        "-if",
        dest="interface",
        help="Interface. Only used if device type can't be autodetected.",
        type=str.lower,
        choices=["usb", "serial"],
    )
    subparser.add_argument(
        "--clear-login", dest="clear", action="store_true", help="Clears saved login session."
    )

    image_group = subparser.add_mutually_exclusive_group(required=False)
    image_group.add_argument("--image", "-i", dest="image", help="Image file to flash")
    image_group.add_argument(
        "--fetch",
        "-f",
        dest="fetch",
        action="store_true",
        help=(
            "Fetch latest image file from the web and flash it. "
            "Requires an Acconeer Developer Account."
        ),
    )
    verbosity_group = parser.add_mutually_exclusive_group(required=False)
    verbosity_group.add_argument(
        "-v",
        "--verbose",
        action="store_true",
    )
    verbosity_group.add_argument(
        "-vv",
        "--debug",
        action="store_true",
    )
    verbosity_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
    )

    args = parser.parse_args()

    et.utils.config_logging(args)

    if args.operation == "list":
        usb_devices = et.utils.get_usb_devices()
        serial_devices = et.utils.get_serial_devices()
        if not usb_devices and not serial_devices:
            print("No devices available")
        else:
            if usb_devices:
                print("== USB Devices ==")
                print("".join([f" - {dev}\n" for dev in usb_devices]))
            if serial_devices:
                print("== Serial Devices ==")
                print("".join([f" - {dev}\n" for dev in serial_devices]))
        return
    elif args.operation == "flash":
        if args.clear:
            clear_cookies()
            print("Saved login session cookie deleted!\n")

        if args.fetch:
            fetch_and_flash(args)
        elif args.image:
            flash_device = get_flash_device_from_args(args, verbose=True)
            flash_image(
                args.image,
                flash_device,
                device_name=args.device,
                progress_callback=flasher_progress_callback,
            )
        elif not args.clear:
            print("No image file selected!\n")
            parser.print_help()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .configs import (
    PRF,
    Criticality,
    IdleState,
    Profile,
    SensorConfig,
    SessionConfig,
    SubsweepConfig,
    ValidationError,
    ValidationResult,
    ValidationWarning,
)
from .containers import (
    ClientInfo,
    Metadata,
    PersistentRecord,
    Record,
    RecordException,
    Result,
    ResultContext,
    SensorCalibration,
    SensorInfo,
    ServerInfo,
    ServerLogMessage,
    StackedResults,
    complex_array_to_int16_complex,
    int16_complex_array_to_complex,
)
from .dtypes import INT_16_COMPLEX



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import TypeVar

import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core.entities.dtypes import INT_16_COMPLEX

from .metadata import Metadata


T = TypeVar("T", bound=np.generic)


def get_subsweeps_from_frame(frame: npt.NDArray[T], metadata: Metadata) -> list[npt.NDArray[T]]:
    """Gets the subsweeps from a frame (2D, (<sweeps>, <data points>))
    or frames (3D, (<frames>, <sweeps>, <data points>)).
    This is done by slicing the innermost dimension (the <data points> dimension) for both cases.
    """
    offsets = metadata.subsweep_data_offset
    lengths = metadata.subsweep_data_length
    return [frame[..., o : o + l] for o, l in zip(offsets, lengths)]


def int16_complex_array_to_complex(array: npt.NDArray) -> npt.NDArray[np.complex_]:
    """Converts an array with dtype = INT_16_COMPLEX
    (structured with parts "real" and "imag") into
    an array with plain complex dtype (non-structured).
    """
    real = array["real"].astype("float")
    imaginary = array["imag"].astype("float")
    return real + 1.0j * imaginary  # type: ignore[no-any-return]


def complex_array_to_int16_complex(array: npt.NDArray[np.complex_]) -> npt.NDArray:
    """Converts an array with plain complex dtype (non-structured)
    into an array with dtype = INT_16_COMPLEX
    (structured with parts "real" and "imag") using `numpy.round`.
    """
    struct_array = np.empty(array.shape, dtype=INT_16_COMPLEX)
    struct_array["real"] = np.round(array.real)
    struct_array["imag"] = np.round(array.imag)
    return struct_array



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from __future__ import annotations

import typing as t

import attrs
import typing_extensions as te

from acconeer.exptool.a121._core.entities import SensorInfo
from acconeer.exptool.a121._core.peripherals.communication.message import Message

from .parse_error import ParseError


class SensorInfoHeader(te.TypedDict):
    connected: bool
    serial: t.Optional[str]


@attrs.frozen
class SensorInfoResponse(Message):
    sensor_infos: t.Dict[int, SensorInfo]

    @classmethod
    def parse(cls, header: t.Dict[str, t.Any], payload: bytes) -> SensorInfoResponse:
        try:
            sensor_infos: t.List[SensorInfoHeader] = header["sensor_info"]

            return cls(
                {
                    i: SensorInfo(
                        connected=sensor_info_dict["connected"],
                        serial=sensor_info_dict.get("serial", None),
                    )
                    for i, sensor_info_dict in enumerate(sensor_infos, start=1)
                }
            )
        except KeyError as ke:
            raise ParseError from ke



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import typing as t

from .app_model import AppModel


class AppModelListener:
    """A listener of AppModel

    Subclasses needs to do `super().__init__(app_model)` in order
    to set up listening.
    """

    def __init__(self, app_model: AppModel) -> None:
        self.__app_model = app_model
        self.__app_model.sig_notify.connect(self.on_app_model_update)
        self.__app_model.sig_backend_state_changed.connect(self.on_backend_state_update)

    def on_app_model_update(self, app_model: AppModel) -> None:
        """Hook for getting updates made to AppModel.

        AppModel calls this function when `AppModel.sig_notify` is emitted,
        i.e. when `AppModel.broadcast` is invoked.
        """
        pass

    def on_backend_state_update(self, backend_plugin_state: t.Optional[t.Any]) -> None:
        """Hook for getting updates made to the backend plugin.

        AppModel calls this function when `AppModel.sig_backend_state_changed` is emitted,
        i.e. when `AppModel.broadcast_backend_state` is invoked.
        """
        pass

    def stop_listening(self) -> None:
        """Stops listening to AppModel"""
        self.__app_model.sig_notify.disconnect(self.on_app_model_update)
        self.__app_model.sig_backend_state_changed.disconnect(self.on_backend_state_update)



# Copyright (c) Acconeer AB, 2023
# All rights reserved
from __future__ import annotations

import copy
import typing as t

import h5py
import numpy as np
import numpy.typing as npt

from acconeer.exptool.a121._core import INT_16_COMPLEX, complex_array_to_int16_complex
from acconeer.exptool.a121._core.entities.containers.utils import int16_complex_array_to_complex

from ._processors import ProcessorResult


S = t.TypeVar("S")
T = t.TypeVar("T")
DTypeT = t.TypeVar("DTypeT")

_ALL_RESULT_FIELDS: t.Final = (
    "estimated_distances",
    "estimated_rcs",
    "recorded_threshold_mean_sweep",
    "recorded_threshold_noise_std",
    "direct_leakage",
    "phase_jitter_comp_reference",
    # "extra_result" ignored by default
)

_INT_16_COMPLEX_SENTINEL = -(2**15)


class PhonySeries(t.Generic[T]):
    def __init__(self, prototype: T, is_prototype_singleton: bool = True) -> None:
        self._prototype = prototype
        self._is_prototype_singleton = is_prototype_singleton

    def __next__(self) -> T:
        if self._is_prototype_singleton:
            return self._prototype
        else:
            return copy.copy(self._prototype)

    def __iter__(self) -> PhonySeries:
        return self


class ProcessorResultListH5Serializer:
    """
    Reads or writes a distance ProcessorResult from/to a given h5py.Group
    """

    def __init__(
        self,
        group: h5py.Group,
        fields: t.Sequence[str] = _ALL_RESULT_FIELDS,
        allow_missing_fields: bool = False,
    ) -> None:
        """
        :param destination: The H5 group that will have datasets read/written to it
        :param fields: The fields to serialize. Has no effect on de-serialization
        :param allow_missing_fields:
            Whether it's acceptable to break type safety during de-serialization
            by replacing missing fields by ``None``. Has no effect on serialization.
        """
        self.group = group
        self.fields = fields
        self.allow_missing_fields = allow_missing_fields

    @staticmethod
    def _stack_optional_arraylike(
        sequence: t.Sequence[t.Optional[npt.ArrayLike]],
        dtype: t.Union[t.Type[float], t.Type[complex]] = float,
    ) -> npt.NDArray:
        """
        Tries to create an NDArray from a sequence of Optional ArrayLikes, i.e.
        a sequence that looks something like

            [[1 2 3],
             None,
             [1 2 3],
             ...
            ]

        ``None`` is replaced with a NaN-array that has the same dimensions as the
        non-NaN array in the sequence.
        """
        SENTINELS = {float: np.NAN, complex: np.NAN + 1j * np.NAN}
        shapes = {np.shape(x) for x in sequence if x is not None}

        if len(shapes) > 1:
            raise ValueError(f"arrays in {sequence} has different lengths.")
        elif len(shapes) == 0:
            raise ValueError("All elements in sequence are None")

        (shape,) = shapes

        return np.stack(
            [
                np.full(shape, fill_value=SENTINELS[dtype], dtype=dtype)
                if x is None
                else np.array(x)
                for x in sequence
            ]
        )

    def serialize(self, results: t.List[ProcessorResult]) -> None:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        if "estimated_distances" in self.fields:
            try:
                data = self._stack_optional_arraylike([r.estimated_distances for r in results])
            except ValueError:
                pass
            else:
                self.group.create_dataset(
                    "estimated_distances",
                    dtype=float,
                    data=data,
                    track_times=False,
                )

        if "estimated_rcs" in self.fields:
            try:
                data = self._stack_optional_arraylike([r.estimated_rcs for r in results])
            except ValueError:
                pass
            else:
                self.group.create_dataset(
                    "estimated_rcs",
                    dtype=float,
                    data=data,
                    track_times=False,
                )

        if "recorded_threshold_mean_sweep" in self.fields:
            try:
                data = self._stack_optional_arraylike(
                    [r.recorded_threshold_mean_sweep for r in results]
                )
            except ValueError:
                pass
            else:
                self.group.create_dataset(
                    "recorded_threshold_mean_sweep",
                    dtype=float,
                    data=data,
                    track_times=False,
                )

        if "recorded_threshold_noise_std" in self.fields:
            try:
                data = self._stack_optional_arraylike(
                    [r.recorded_threshold_noise_std for r in results]
                )
            except ValueError:
                pass
            else:
                self.group.create_dataset(
                    "recorded_threshold_noise_std",
                    dtype=float,
                    data=data,
                    track_times=False,
                )

        if "direct_leakage" in self.fields:
            try:
                data = self._stack_optional_arraylike(
                    [r.direct_leakage for r in results], dtype=complex
                )
            except ValueError:
                pass
            else:
                # replaces NaNs with int-sentinels before int_16_complex-conversion
                # since NaN becomes 0 at conversion
                data = np.array(
                    [
                        np.full(x.shape, fill_value=_INT_16_COMPLEX_SENTINEL * (1 + 1j))
                        if np.isnan(x).all()
                        else x
                        for x in data
                    ]
                )

                self.group.create_dataset(
                    "direct_leakage",
                    dtype=INT_16_COMPLEX,
                    data=complex_array_to_int16_complex(data),
                    track_times=False,
                )

        if "phase_jitter_comp_reference" in self.fields:
            try:
                data = self._stack_optional_arraylike(
                    [res.phase_jitter_comp_reference for res in results]
                )
            except ValueError:
                pass
            else:
                self.group.create_dataset(
                    "phase_jitter_comp_reference",
                    dtype=float,
                    data=data,
                    track_times=False,
                )

    @staticmethod
    def _replace_if_all_is_nan(
        x: t.Optional[npt.ArrayLike], replacement: t.Any = None
    ) -> t.Union[npt.ArrayLike, t.Any]:
        if x is None:
            return replacement

        return replacement if np.isnan(x).all() else x

    @staticmethod
    def _direct_leakage_deserialize(x: t.Optional[npt.NDArray]) -> t.Optional[npt.NDArray]:
        if x is None:
            return None
        elif (x == np.array(-(2**15), dtype=INT_16_COMPLEX)).all():
            return None
        else:
            return int16_complex_array_to_complex(x)

    def deserialize(self, _: None) -> t.List[ProcessorResult]:
        if "extra_result" in self.fields:
            raise NotImplementedError(
                "'extra_results' are not serializable at the moment."
                + "Skip it by specifying which fields to serialize"
            )

        groups = (
            self.group.get("estimated_distances", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("estimated_rcs", PhonySeries([], is_prototype_singleton=False)),
            self.group.get("recorded_threshold_mean_sweep", PhonySeries(None)),
            self.group.get("recorded_threshold_noise_std", PhonySeries(None)),
            self.group.get("direct_leakage", PhonySeries(None)),
            self.group.get("phase_jitter_comp_reference", PhonySeries(None)),
        )

        if any(isinstance(g, PhonySeries) for g in groups) and not self.allow_missing_fields:
            raise ValueError("Some fields are missing while 'allow_missing_fields' is False.")

        if all(isinstance(group, PhonySeries) for group in groups):
            return []

        return [
            ProcessorResult(
                estimated_distances=self._replace_if_all_is_nan(
                    est_dists,
                ),  # type: ignore[arg-type]
                estimated_rcs=self._replace_if_all_is_nan(
                    est_rcs,
                ),  # type: ignore[arg-type]
                recorded_threshold_mean_sweep=(
                    self._replace_if_all_is_nan(rt_mean_sweep)  # type: ignore[arg-type]
                ),
                recorded_threshold_noise_std=(
                    self._replace_if_all_is_nan(rt_noise_std)  # type: ignore[arg-type]
                ),
                direct_leakage=(self._direct_leakage_deserialize(direct_leakage)),
                phase_jitter_comp_reference=(
                    self._replace_if_all_is_nan(jitter_comp)  # type: ignore[arg-type]
                ),
                extra_result=None,  # type: ignore[arg-type]
            )
            for (
                est_dists,
                est_rcs,
                rt_mean_sweep,
                rt_noise_std,
                direct_leakage,
                jitter_comp,
            ) in zip(*groups)
        ]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from enum import Enum, auto
from typing import Callable, Type

import numpy as np

from PySide6.QtWidgets import QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
)
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo.vibration import (
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    pidgets,
)


class PluginPresetId(Enum):
    DEFAULT = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):

    PLUGIN_PRESETS = {
        PluginPresetId.DEFAULT.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_sensor_config()),
            processor_config=BackendPlugin.get_processor_config_cls()(),
        ),
    }

    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "time_series_length": pidgets.IntParameterWidgetFactory(
                name_label_text="Time series length",
                limits=(0, None),
            ),
            "lp_coeff": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Time filtering coefficient",
                suffix="",
                limits=(0, 1),
                decimals=2,
            ),
            "sensitivity": pidgets.FloatSliderParameterWidgetFactory(
                name_label_text="Threshold sensitivity",
                suffix="dB",
                limits=(0, 30),
                decimals=1,
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:

        self.meas_dist_m = sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        brush_dot = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush_dot, symbolPen="k")

        # precense plot
        self.precense_plot = pg.PlotItem()
        self.precense_plot.setMenuEnabled(False)
        self.precense_plot.showGrid(x=False, y=True)
        self.precense_plot.setLabel("left", "Max amplitude")
        self.precense_plot.setLabel("bottom", "Distance (m)")
        self.precense_plot.setXRange(self.meas_dist_m - 0.001, self.meas_dist_m + 0.001)
        self.precense_curve = self.precense_plot.plot(**dict(pen=pen, **symbol_dot_kw))

        self.smooth_max_precense = et.utils.SmoothMax(tau_decay=10.0)

        # sweep and threshold plot
        self.time_series_plot = pg.PlotItem()
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.setLabel("left", "Displacement (mm)")
        self.time_series_plot.setLabel("bottom", "History")
        self.time_series_curve = self.time_series_plot.plot(**feat_kw)

        sublayout = self.plot_layout.addLayout(row=0, col=0)
        sublayout.layout.setColumnStretchFactor(1, 5)
        sublayout.addItem(self.precense_plot, row=0, col=0)
        sublayout.addItem(self.time_series_plot, row=0, col=1)

        self.smooth_lim_time_series = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

        self.fft_plot = self.plot_layout.addPlot(col=0, row=1)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.setLabel("left", "Power (dB)")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = [
            self.fft_plot.plot(**feat_kw),
            self.fft_plot.plot(**dict(pen=pen, **symbol_dot_kw)),
        ]

        self.text_item = pg.TextItem(
            fill=pg.mkColor(0xFF, 0x7F, 0x0E, 200),
            anchor=(0.5, 0),
        )
        self.text_item.hide()
        self.fft_plot.addItem(self.text_item)

        self.smooth_max_fft = et.utils.SmoothMax()

    def update(self, processor_result: ProcessorResult) -> None:

        time_series = processor_result.time_series
        z_abs_db = processor_result.lp_z_abs_db
        freqs = processor_result.freqs
        max_amplitude = processor_result.max_amplitude
        max_psd_ampl = processor_result.max_psd_ampl
        max_psd_ampl_freq = processor_result.max_psd_ampl_freq

        self.precense_curve.setData([self.meas_dist_m], [max_amplitude])
        lim = self.smooth_max_precense.update(max_amplitude)
        self.precense_plot.setYRange(0, max(1000.0, lim))

        self.time_series_curve.setData(time_series)
        lim = self.smooth_lim_time_series.update(time_series)
        self.time_series_plot.setYRange(lim[0], lim[1])

        self.fft_curve[0].setData(freqs, z_abs_db)
        lim = self.smooth_max_fft.update(np.max(z_abs_db))
        self.fft_plot.setYRange(0, lim)

        if max_psd_ampl_freq is not None:
            self.fft_curve[1].setData([max_psd_ampl_freq], [max_psd_ampl])
            # Place text box centered at the top of the plotting window.
            self.text_item.setPos(max(freqs) / 2, lim * 0.95)
            html_format = (
                '<div style="text-align: center">'
                '<span style="color: #FFFFFF;font-size:15pt;">'
                "{}</span></div>".format("Detected Frequency: " + str(int(max_psd_ampl_freq)))
            )
            self.text_item.setHtml(html_format)
            self.text_item.show()
        else:
            self.fft_curve[1].setData([], [])
            self.text_item.hide()


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


VIBRATION_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="vibration",
    title="Vibration measurement",
    description="Quantify the frequency content of vibrating object.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(name="Default", preset_id=PluginPresetId.DEFAULT),
    ],
    default_preset_id=PluginPresetId.DEFAULT,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .plot_plugin_base import PlotPluginBase
from .plugin_preset_base import PluginPresetBase
from .plugin_spec_base import PluginSpecBase
from .view_plugin_base import ViewPluginBase



# Copyright (c) Acconeer AB, 2022
# All rights reserved
from typing import Any

import numpy as np

from ._processor import ProcessorResult


class Easing:
    def __init__(self, tc: float = 0.05, y0: float = 0.0) -> None:
        self.tc = tc
        self.y = y0

    def update(self, x: float, dt: float) -> float:
        d = x - self.y
        self.y += d * (1.0 - np.exp(-(dt / self.tc)))
        return self.y


class BlinkstickUpdater:
    def __init__(self) -> None:
        self.e_close = Easing()
        self.e_far = Easing()
        self.last_t = -1

    def update(self, data: ProcessorResult, data_index: int, t: int, stick: Any) -> None:
        dt = t - self.last_t
        self.last_t = t

        stick.set_max_rgb_value(150)

        v_close = self.e_close.update(float(bool(data.detection_close)), dt)
        v_far = self.e_far.update(float(bool(data.detection_far)), dt)

        stick.set_color(blue=255 * v_far, index=0)
        stick.set_color(blue=255 * v_far, index=1)
        stick.set_color(red=255 * v_close, index=2)
        stick.set_color(red=255 * v_close, index=3)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import json
from typing import Any, Optional

import attrs

from acconeer.exptool.a121._core.utils import pretty_dict_line_strs
from acconeer.exptool.utils import USBDevice  # type: ignore[import]


@attrs.frozen(kw_only=True)
class ClientInfo:
    ip_address: Optional[str] = None
    serial_port: Optional[str] = None
    usb_device: Optional[USBDevice] = None
    mock: Optional[bool] = None
    override_baudrate: Optional[int] = None

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> ClientInfo:
        if d.get("usb_device") is not None:
            d = d.copy()
            d["usb_device"] = USBDevice.from_dict(d["usb_device"])
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> ClientInfo:
        return cls.from_dict(json.loads(json_str))

    def __str__(self) -> str:
        return "\n".join(
            [
                f"{type(self).__name__}",
                *pretty_dict_line_strs(self.to_dict()),
            ]
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import acconeer.exptool as et
from acconeer.exptool import a121


parser = a121.ExampleArgumentParser()
parser.add_argument("--output-file", required=True)
parser.add_argument("--num-frames", required=True, type=int)
args = parser.parse_args()
et.utils.config_logging(args)

# Here we create a H5Recorder.
# The H5Recorder is an object that saves frames directly to a H5-file.
h5_recorder = a121.H5Recorder(args.output_file)

client = a121.Client(**a121.get_client_args(args))
client.connect()

# Session setup, just like the other examples.
config = a121.SessionConfig()
client.setup_session(config)

# Here we send the H5Recorder to the Client.
# After this call, the client is responsible for the H5Recorder.
# The Client will automatically close the file when
# "stop_session" is called.
client.start_session(recorder=h5_recorder)

for i in range(args.num_frames):
    # Client will send its Results to the H5Recorder.
    client.get_next()
    print(f"Result {i + 1}/{args.num_frames} was sampled")

client.stop_session()
client.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional, Union

from PySide6 import QtCore
from PySide6.QtWidgets import (
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool import a121
from acconeer.exptool.a121 import _core

from .utils import VerticalGroupBox


_WIDGET_WIDTH = 125


class SmartMetadataView(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self._layout = QVBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self.setLayout(self._layout)

        self._metadata_view = MetadataView(self)
        self._extended_metadata_view = ExtendedMetadataView(self)
        self._layout.addWidget(self._metadata_view)
        self._layout.addWidget(self._extended_metadata_view)

    def update(
        self,
        metadata: Optional[Union[a121.Metadata, list[dict[int, a121.Metadata]]]] = None,
    ) -> None:
        if isinstance(metadata, list):
            self._metadata_view.setHidden(True)
            self._extended_metadata_view.setHidden(False)
            self._extended_metadata_view.update(metadata)
        else:
            self._metadata_view.setHidden(False)
            self._extended_metadata_view.setHidden(True)
            self._metadata_view.update(metadata)


class ExtendedMetadataView(QWidget):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setLayout(QVBoxLayout(self))

        group_box = VerticalGroupBox("Metadata", parent=self)
        self.layout().addWidget(group_box)

        self._tab_widget = QTabWidget(parent=self)
        self._tab_widget.setStyleSheet(
            """
            QTabWidget::pane { border: none; }
            * {margin: 0;}
            """
        )

        group_box.layout().addWidget(self._tab_widget)
        self._represent_none()

    def _represent_none(self) -> None:
        self._tab_widget.clear()
        self._add_needed_tabs(1)

    def _add_needed_tabs(self, needed_widgets: int) -> None:
        while self._tab_widget.count() < needed_widgets:
            subwidget = MetadataView(self)
            subwidget.setTitle("")
            self._tab_widget.addTab(subwidget, "-")

    def _remove_unnecessary_tabs(self, needed_widgets: int) -> None:
        while self._tab_widget.count() > needed_widgets:
            self._tab_widget.removeTab(0)

    def update(self, extended_metadata: Optional[list[dict[int, a121.Metadata]]]) -> None:
        if extended_metadata is None:
            self._represent_none()
            return

        tabs_needed = _core.utils.extended_structure_entry_count(extended_metadata)
        self._add_needed_tabs(tabs_needed)
        self._remove_unnecessary_tabs(tabs_needed)

        for i, (group_id, sensor_id, metadata) in enumerate(
            _core.utils.iterate_extended_structure(extended_metadata)
        ):
            self._tab_widget.setTabText(i, f"G{group_id}:S{sensor_id}")
            metadata_view = self._tab_widget.widget(i)
            if isinstance(metadata_view, MetadataView):
                metadata_view.update(metadata)
            else:
                raise RuntimeError(
                    "ExtendedMetadataView contains child widgets that are not MetadataViews."
                )


class MetadataView(QGroupBox):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setTitle("Metadata")
        self.setLayout(QGridLayout(self))

        row = 0

        self.frame_data_length = MetadataValueWidget(self)
        self.layout().addWidget(self.frame_data_length, row, 1)

        frame_data_length_label = QLabel("Frame data length", self)
        frame_data_length_label.setToolTip(a121.Metadata.frame_data_length.__doc__)
        self.layout().addWidget(frame_data_length_label, row, 0)

        row += 1

        self.calibration_temperature = MetadataValueWidget(self)
        self.layout().addWidget(self.calibration_temperature, row, 1)

        calibration_temperature_label = QLabel("Calibration temperature", self)
        calibration_temperature_label.setToolTip(a121.Metadata.calibration_temperature.__doc__)
        self.layout().addWidget(calibration_temperature_label, row, 0)

        row += 1

        self.max_sweep_rate = MetadataValueWidget(self)
        self.layout().addWidget(self.max_sweep_rate, row, 1)

        max_sweep_rate_label = QLabel("Max sweep rate", self)
        max_sweep_rate_label.setToolTip(a121.Metadata.max_sweep_rate.__doc__)
        self.layout().addWidget(max_sweep_rate_label, row, 0)

        self.update(None)

    def update(self, metadata: Optional[a121.Metadata]) -> None:
        self.frame_data_length.setText(f"{metadata.frame_data_length}" if metadata else "-")
        self.calibration_temperature.setText(
            f"{metadata.calibration_temperature}" if metadata else "-"
        )

        if metadata:
            if metadata.max_sweep_rate:
                max_sweep_rate_text = f"{metadata.max_sweep_rate:.0f} Hz"
            else:
                max_sweep_rate_text = "N/A"
        else:
            max_sweep_rate_text = "-"

        self.max_sweep_rate.setText(max_sweep_rate_text)


class MetadataValueWidget(QLineEdit):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)
        self.setAlignment(QtCore.Qt.AlignRight)
        self.setFixedWidth(_WIDGET_WIDTH)
        self.setReadOnly(True)
        self.setFocusPolicy(QtCore.Qt.NoFocus)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import time
import typing as t

import acconeer.exptool as et
from acconeer.exptool import a121


def main() -> None:
    parser = a121.ExampleArgumentParser()

    parser.add_argument(
        "program",
        choices=[
            "throughput",
            "rate",
            "short_pauses",
            "long_pauses",
            "flow",
        ],
    )

    args = parser.parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()
    print(client.client_info)
    print()
    print(client.server_info)
    print()

    if args.program == "throughput":
        func = throughput
    elif args.program == "rate":
        func = rate
    elif args.program == "short_pauses":
        func = short_pauses
    elif args.program == "long_pauses":
        func = long_pauses
    elif args.program == "flow":
        func = flow
    else:
        raise RuntimeError

    try:
        func(client)
    except KeyboardInterrupt:
        print()
        try:
            client.disconnect()
        except Exception:
            pass


def flow(client: a121.Client) -> None:
    client.disconnect()

    n = 0

    while True:
        client.connect()

        for _ in range(5):
            client.setup_session(a121.SensorConfig())
            client.start_session()

            for _ in range(5):
                client.get_next()

            client.stop_session()

        client.disconnect()

        n += 1

        print_status(f"{n:8}")


def throughput(client: a121.Client) -> None:
    config = a121.SensorConfig(
        hwaas=1,
        profile=a121.Profile.PROFILE_5,
        prf=a121.PRF.PRF_13_0_MHz,
        sweeps_per_frame=1,
        start_point=0,
        num_points=2047,
        step_length=1,
        double_buffering=True,
    )
    _stress_session(client, config)


def rate(client: a121.Client) -> None:
    config = a121.SensorConfig(
        hwaas=1,
        profile=a121.Profile.PROFILE_5,
        prf=a121.PRF.PRF_13_0_MHz,
        sweeps_per_frame=1,
        start_point=0,
        num_points=1,
        step_length=1,
        double_buffering=True,
    )
    _stress_session(client, config)


def short_pauses(client: a121.Client) -> None:
    _pauses(client, pause_time=0.25)


def long_pauses(client: a121.Client) -> None:
    _pauses(client, pause_time=1.5)


def _pauses(client: a121.Client, *, pause_time: float) -> None:
    config = a121.SensorConfig(
        hwaas=1,
        profile=a121.Profile.PROFILE_5,
        prf=a121.PRF.PRF_13_0_MHz,
        sweeps_per_frame=1,
        start_point=0,
        num_points=500,
        step_length=1,
        frame_rate=100.0,
    )
    _stress_session(client, config, pause_time=pause_time)


def _stress_session(
    client: a121.Client,
    config: a121.SensorConfig,
    *,
    pause_time: t.Optional[float] = None,
) -> None:
    if pause_time is None:
        pause_interval = None
    else:
        pause_interval = pause_time * 2

    client.setup_session(config)
    client.start_session()

    n = 0
    last_pause = time.monotonic()

    while True:
        client.get_next()

        n += 1

        now = time.monotonic()
        if pause_interval is not None and (now - last_pause) > pause_interval:
            last_pause = now
            assert pause_time is not None
            time.sleep(pause_time)

        stats = client._rate_stats

        bits_per_point = 32
        points_per_frame = config.num_points * config.sweeps_per_frame
        effective_bitrate = stats.rate * points_per_frame * bits_per_point

        print_status(
            " | ".join(
                [
                    f"{n:8}",
                    f"{stats.rate:7.1f} Hz",
                    f"{stats.jitter * 1e6:7.1f} us jitter",
                    f"{effective_bitrate * 1e-6:7.1f} Mbit/s",
                ]
            )
        )


def print_status(s: str) -> None:
    now = time.monotonic()
    last_time = getattr(print_status, "last_time", -1)
    dt = now - last_time

    if dt < 0.25:
        return

    setattr(print_status, "last_time", now)
    print(s, end="\r", flush=True)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import cgi
import logging
import os
import pickle
import shutil
import tempfile
import zipfile
from pathlib import Path
from typing import Optional, Tuple

import requests

import platformdirs
from bs4 import BeautifulSoup


ET_DIR = Path(platformdirs.user_data_dir(appname="acconeer_exptool", appauthor="Acconeer AB"))
CODENAME = "plugoneer"
COOKIE_DIR = ET_DIR / CODENAME / "cookies"
COOKIE_FILEPATH = COOKIE_DIR / "developer_page.pickle"
REQUEST_URL = "https://developer.acconeer.com/log-in/"
REFERER_URL = "https://developer.acconeer.com"
ACC_DEV_AJAX_URL = "https://developer.acconeer.com/wp-admin/admin-ajax.php"

BIN_FETCH_PROMPT = (
    "To fetch the latest image you need to log into your Acconeer Developer account.\n\n"
    "If you don't have an account you can create one here:\n"
    "https://developer.acconeer.com/create-account/ \n\n"
    "Lost your password? Recover it here:\n"
    "https://developer.acconeer.com/recover-password/"
)

log = logging.getLogger(__name__)


def get_cookies() -> requests.cookies.RequestsCookieJar:
    if os.path.exists(COOKIE_FILEPATH):
        with open(COOKIE_FILEPATH, "rb") as handle:
            cookie_jar = pickle.load(handle)
        cookie_jar.clear_expired_cookies()
        cookies = cookie_jar.get_dict()
        if not cookies:
            cookie_jar = None
    else:
        cookie_jar = None
    return cookie_jar


def _make_cookie_dir() -> None:
    COOKIE_DIR.mkdir(parents=True, exist_ok=True)


def save_cookies(cookies: requests.cookies.RequestsCookieJar) -> None:
    _make_cookie_dir()
    with open(COOKIE_FILEPATH, "wb") as handle:
        pickle.dump(cookies, handle)
    log.debug("Login saved")


def clear_cookies() -> None:
    if os.path.exists(COOKIE_FILEPATH):
        os.remove(COOKIE_FILEPATH)
        log.debug("Login session removed")
    else:
        log.debug("No previous session saved")


def login(email: str, password: str) -> requests.cookies.RequestsCookieJar:

    with requests.Session() as session:
        login_page = session.get(REQUEST_URL)
        soup = BeautifulSoup(login_page.content, "html.parser")
        token = soup.find("input", {"id": "CSRFToken-wppb"}).get("value")

        credentials = {
            "log": email,
            "pwd": password,
            "rememberme": "forever",
            "wp-submit": "Log+In",
            "wppb_login": "true",
            "wppb_request_url": REQUEST_URL,
            "wppb_redirect_priority": "normal",
            "wppb_referer_url": REFERER_URL,
            "CSRFToken-wppb": token,
            "_wp_http_referer": "/log-in/",
            "wppb_redirect_check": "true",
        }
        sw_page = session.post(REQUEST_URL, data=credentials, allow_redirects=False)
        cookies = sw_page.cookies

        return cookies


def get_content(
    cookies: requests.cookies.RequestsCookieJar,
) -> Tuple[bool, requests.Session, requests.Response]:

    with requests.Session() as session:
        response = session.get(url=REFERER_URL, cookies=cookies.get_dict())

        if response.url == REQUEST_URL:
            return False, session, response
        elif response.headers.get("Expires") is None:
            return False, session, response
        elif response.headers.get("Expires") is not None:
            return True, session, response
        else:
            raise Exception


def is_redirected_to_login_page(response: requests.Response) -> bool:
    return response.url == REQUEST_URL


def download(
    session: requests.Session,
    cookies: dict,
    path: str,
    device: str,
) -> Tuple[str, str]:

    device = device.lower()

    # Get correct device slug
    slug = "xe121" if device == "xc120" else device

    data = f"action=ajax_file_tabs&slug={slug}"
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    request = session.post(url=ACC_DEV_AJAX_URL, data=data, cookies=cookies, headers=headers)

    soup = BeautifulSoup(request.content, "html.parser")
    form = soup.select_one('form[action*="{}"]'.format(device))
    if form is None:
        raise Exception(f"No image found for device '{device}'")
    zip_url = form.get("action")

    log.debug("File to download: {}".format(zip_url))

    tc = "tc_accepted=1"
    headers = {"Content-Type": "application/x-www-form-urlencoded"}

    r = session.post(zip_url, data=tc, allow_redirects=False, cookies=cookies, headers=headers)

    _, params = cgi.parse_header(r.headers["Content-Disposition"])
    filename = params["filename"]

    version = _get_version(filename)

    filepath = os.path.join(path, filename)

    with open(filepath, "wb") as output_file:
        output_file.write(r.content)
        log.debug("Download completed at {}".format(filepath))

    log.debug("Extracting files..")

    with tempfile.TemporaryDirectory() as tmp_dir:
        bin_found = False
        with zipfile.ZipFile(filepath, "r") as zipObject:
            listOfFileNames = zipObject.namelist()
            for fileName in listOfFileNames:
                if fileName.endswith(".bin"):
                    if bin_found:
                        raise Exception("Multiple images in directory")
                    bin_path = zipObject.extract(fileName, os.path.join(path, tmp_dir))
                    bin_found = True
            if bin_found is False:
                raise Exception("No images found in directory")

        dst_path = os.path.join(path, os.path.basename(bin_path))
        shutil.move(bin_path, dst_path)

        os.remove(filepath)

    return dst_path, version


def _get_version(bin_file: Optional[str]) -> str:
    version = "n/a"

    if bin_file is not None:
        version = Path(bin_file).stem

    return version



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from typing_extensions import Protocol

from acconeer.exptool.app.new.backend import GeneralMessage


class _MessageHandler(Protocol):
    def handle_message(self, message: GeneralMessage) -> None:
        ...


class ViewPluginInterface(_MessageHandler, Protocol):
    pass


class PlotPluginInterface(_MessageHandler, Protocol):
    def draw(self) -> None:
        ...



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
from contextlib import contextmanager
from typing import Any, Callable, Generic, Optional, TypeVar

import h5py

from acconeer.exptool.app.new import PluginGeneration
from acconeer.exptool.app.new.storage import get_config_dir

from ._message import BackendPluginStateMessage, GeneralMessage, Message, StatusMessage


StateT = TypeVar("StateT")


class BackendPlugin(abc.ABC, Generic[StateT]):
    shared_state: StateT

    def __init__(
        self, callback: Callable[[Message], None], generation: PluginGeneration, key: str
    ) -> None:
        self.callback = callback
        self.key = key
        self.generation = generation

    @contextmanager
    def h5_cache_file(self, write: bool = False) -> h5py.File:
        file_path = (get_config_dir() / "plugin" / self.generation.value / self.key).with_suffix(
            ".h5"
        )

        if write:
            file_path.parent.mkdir(parents=True, exist_ok=True)

        file_mode = "w" if write else "r"
        h5_file = h5py.File(file_path, file_mode)
        yield h5_file
        h5_file.close()

    @abc.abstractmethod
    def load_from_cache(self) -> None:
        pass

    @abc.abstractmethod
    def idle(self) -> bool:
        pass

    @abc.abstractmethod
    def attach_client(self, *, client: Any) -> None:
        pass

    @abc.abstractmethod
    def detach_client(self) -> None:
        pass

    @abc.abstractmethod
    def teardown(self) -> None:
        pass

    @abc.abstractmethod
    def set_preset(self, preset_id: int) -> None:
        pass

    def broadcast(self, sync: bool = False) -> None:
        self.callback(BackendPluginStateMessage(state=self.shared_state))

        if sync:
            self.callback(GeneralMessage(name="sync", recipient="view_plugin"))

    def send_status_message(self, message: Optional[str]) -> None:
        self.callback(StatusMessage(status=message))



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .app_model import AppModel, PluginPresetSpec, PluginSpec
from .app_model_listener import AppModelListener
from .plugin_protocols import PlotPluginInterface, ViewPluginInterface



# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import PERCEIVED_WAVELENGTH, AlgoProcessorConfigBase, ProcessorBase


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    threshold: Optional[float] = attrs.field(default=25.0)

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweep_rate",
                    "Must be set",
                )
            )

        if not config.sensor_config.double_buffering:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "double_buffering",
                    "Must be enabled",
                )
            )

        if not config.sensor_config.continuous_sweep_mode:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "continuous_sweep_mode",
                    "Must be enabled",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext:
    ...


@attrs.frozen(kw_only=True)
class ProcessorResult:
    lp_abs_sweep: npt.NDArray[np.float_]
    angle_sweep: npt.NDArray[np.float_]
    threshold: Optional[float] = attrs.field(default=None)
    rel_time_stamps: npt.NDArray[np.float_]
    distance_history: npt.NDArray[np.float_]
    peak_loc_m: Optional[float] = attrs.field(default=None)


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    estimated_distance: Optional[float]
    prev_peak_loc_m: Optional[float]

    M_TO_MM = 1000
    TIME_HORIZON_S = 10.0
    LP_COEFF = 0.75

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:

        processor_config.validate(sensor_config)

        # Should never happen, checked in validate
        assert sensor_config.sweep_rate is not None

        self.start_point = sensor_config.start_point
        self.step_length = sensor_config.step_length
        self.num_points = sensor_config.num_points

        self.base_step_length_m = metadata.base_step_length_m

        self.prev_peak_loc_m = None

        self.lp_abs_sweep = np.full(self.num_points, np.nan)

        self.sweep_index = 0

        self.threshold = processor_config.threshold

        self.max_num_points_to_plot = int(sensor_config.sweep_rate * self.TIME_HORIZON_S)
        self.distance_history: npt.NDArray[np.float_] = np.array([])

    def process(self, result: a121.Result) -> ProcessorResult:

        frame = result.frame
        abs_sweep = np.mean(np.abs(frame), axis=0)

        # initialize filter variables if first sweep
        if self.sweep_index == 0:
            self.lp_abs_sweep = abs_sweep

        self.lp_abs_sweep = self.lp_abs_sweep * self.LP_COEFF + abs_sweep * (1 - self.LP_COEFF)

        if self.threshold < np.max(self.lp_abs_sweep):
            peak_loc_p = np.argmax(self.lp_abs_sweep)
            peak_loc_m = float(
                (self.start_point + peak_loc_p * self.step_length) * self.base_step_length_m
            )
            # Assign prev_peak_loc_m a value if this is the first sweep above the threshold.
            if self.prev_peak_loc_m is None:
                self.prev_peak_loc_m = peak_loc_m

            # Reset estimate if this is the first amplitude above the threshold(length of
            # distance_history is 0) or distance between the current and previous peak location
            # is large, indicating new object with greater peak.
            if self.distance_history.shape[0] == 0 or 0.1 < np.abs(
                peak_loc_m - self.prev_peak_loc_m
            ):
                self.distance_history = np.array([0.0])

            delta_angles = np.diff(np.unwrap(np.angle(frame[:, peak_loc_p])))
            delta_dists = PERCEIVED_WAVELENGTH * delta_angles / (2 * np.pi) * self.M_TO_MM

            # Append the new distances to the previous distances. Offset the new values with
            # the last value in the existing series for a smooth transition.
            self.distance_history = np.append(
                self.distance_history, self.distance_history[-1] + np.cumsum(delta_dists)
            )

            # Extract data in desired plot window.
            distance_to_plot = self.distance_history[-self.max_num_points_to_plot :]
            time_series_length = distance_to_plot.shape[0]
            rel_time_to_plot = np.linspace(
                -self.TIME_HORIZON_S * time_series_length / self.max_num_points_to_plot,
                0,
                time_series_length,
            )
        else:
            # Reset variables as no peak is detected.
            peak_loc_m = None
            distance_to_plot = np.array([])
            self.prev_peak_loc_m = None
            rel_time_to_plot = np.array([])
            self.distance_history = np.array([])

        self.sweep_index += 1
        self.prev_peak_loc_m = peak_loc_m

        return ProcessorResult(
            lp_abs_sweep=self.lp_abs_sweep,
            angle_sweep=np.angle(np.mean(frame, axis=0)),
            threshold=self.threshold,
            rel_time_stamps=rel_time_to_plot,
            distance_history=distance_to_plot,
            peak_loc_m=peak_loc_m,
        )

    def update_config(self, config: ProcessorConfig) -> None:
        ...


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        profile=a121.Profile.PROFILE_3,
        start_point=120,
        num_points=70,
        step_length=4,
        receiver_gain=10,
        hwaas=4,
        phase_enhancement=True,
        continuous_sweep_mode=True,
        sweeps_per_frame=25,
        sweep_rate=500,
        double_buffering=True,
        inter_sweep_idle_state=a121.IdleState.READY,
        inter_frame_idle_state=a121.IdleState.READY,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved


class PassthroughProcessor:
    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        pass

    def process(self, data, data_info):
        return data


class CompositeProcessor:
    def __init__(
        self, sensor_config, processing_config, session_info, processor_class, calibration=None
    ):
        self.processor_class = processor_class
        self.child_processors = []
        for _ in sensor_config.sensor:
            p = self.processor_class(sensor_config, processing_config, session_info)
            self.child_processors.append(p)

    def update_processing_config(self, processing_config):
        if hasattr(self.processor_class, "update_processing_config"):
            for p in self.child_processors:
                p.update_processing_config(processing_config)

    def process(self, data, data_info):
        return [p.process(d, i) for p, d, i in zip(self.child_processors, data, data_info)]


class MultiSensorProcessorCreator:
    def __init__(self, processor_class):
        self.processor_class = processor_class

    def create_processor(self, sensor_config, processing_config, session_info, calibration=None):
        return CompositeProcessor(
            sensor_config, processing_config, session_info, self.processor_class, calibration
        )


class CompositePGUpdater:
    def __init__(self, sensor_config, processing_config, session_info, updater_class):
        self.updater_class = updater_class
        self.child_updaters = []
        for _ in sensor_config.sensor:
            u = self.updater_class(sensor_config, processing_config, session_info)
            self.child_updaters.append(u)

    def update_processing_config(self, processing_config):
        if hasattr(self.updater_class, "update_processing_config"):
            for u in self.child_updaters:
                u.update_processing_config(processing_config)

    def setup(self, win):
        for i, u in enumerate(self.child_updaters):
            sublayout = win.addLayout(row=0, col=i)
            u.setup(sublayout)

    def update(self, data):
        for u, d in zip(self.child_updaters, data):
            u.update(d)


class MultiSensorPGUpdaterCreator:
    def __init__(self, updater_class):
        self.updater_class = updater_class

    def create_pg_updater(self, sensor_config, processing_config, session_info):
        return CompositePGUpdater(
            sensor_config, processing_config, session_info, self.updater_class
        )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional, Type

from packaging.version import Version

from ._latest import ExplorationProtocol, ExplorationProtocolError
from ._no_5_2_mhz import ExplorationProtocol_No_5_2MHz_PRF
from ._no_15_6_mhz import ExplorationProtocol_No_15_6MHz_PRF
from ._no_calibration_reuse import ExplorationProtocol_NoCalibrationReuse


def get_exploration_protocol(rss_version: Optional[Version] = None) -> Type[ExplorationProtocol]:
    if rss_version is None:
        return ExplorationProtocol

    if rss_version <= Version("0.2.0"):
        raise ExplorationProtocolError("Unsupported RSS version")

    if rss_version < Version("0.4.3.dev280"):
        return ExplorationProtocol_NoCalibrationReuse

    if rss_version < Version("0.4.3.dev310"):
        return ExplorationProtocol_No_5_2MHz_PRF

    if rss_version < Version("0.7.1.dev37"):
        return ExplorationProtocol_No_15_6MHz_PRF

    return ExplorationProtocol



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional, Union

from typing_extensions import Protocol

from acconeer.exptool.a121._core.entities import (
    ClientInfo,
    Metadata,
    Result,
    SensorCalibration,
    SensorConfig,
    ServerInfo,
    SessionConfig,
)
from acconeer.exptool.a121._rate_calc import _RateStats

from .recorder import Recorder


class ClientError(Exception):
    pass


class ClientBase(Protocol):
    def connect(self) -> None:
        """Connects to the specified host.

        :raises: Exception if the host cannot be connected to.
        :raises: ClientError if server has wrong sensor generation (e.g. "a111")
        """
        ...

    def setup_session(
        self,
        config: Union[SensorConfig, SessionConfig],
        calibrations: Optional[dict[int, SensorCalibration]] = None,
    ) -> Union[Metadata, list[dict[int, Metadata]]]:
        """Sets up the session specified by ``config``.

        If the Client is not already connected, it will connect before setting up the session.

        :param config: The session to set up.
        :param calibrations: An optional dict with :class:`SensorCalibration` for the session.
        :raises:
            ``ValueError`` if the config is invalid.

        :returns:
            ``Metadata`` if ``config.extended is False``,
            ``list[dict[int, Metadata]]`` otherwise.
        """
        ...

    def start_session(self, recorder: Optional[Recorder] = None) -> None:
        """Starts the already set up session.

        After this call, the server starts streaming data to the client.

        :param recorder:
            An optional ``Recorder``, which samples every ``get_next()``
        :raises: ``ClientError`` if ``Client``'s  session is not set up.
        """
        ...

    def get_next(self) -> Union[Result, list[dict[int, Result]]]:
        """Gets results from the server.

        :returns:
            A ``Result`` if the setup ``SessionConfig.extended is False``,
            ``list[dict[int, Result]]`` otherwise.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    def stop_session(self) -> Any:
        """Stops an on-going session

        :returns:
            The return value of the passed ``Recorder.stop()`` passed in ``start_session``.
        :raises:
            ``ClientError`` if ``Client``'s session is not started.
        """
        ...

    def disconnect(self) -> None:
        """Disconnects the client from the host."""
        ...

    @property
    def connected(self) -> bool:
        """Whether this Client is connected."""
        ...

    @property
    def session_is_setup(self) -> bool:
        """Whether this Client has a session set up."""
        ...

    @property
    def session_is_started(self) -> bool:
        """Whether this Client's session is started."""
        ...

    @property
    def server_info(self) -> ServerInfo:
        """The ``ServerInfo``."""
        ...

    @property
    def client_info(self) -> ClientInfo:
        """The ``ClientInfo``."""
        ...

    @property
    def session_config(self) -> SessionConfig:
        """The :class:`SessionConfig` for the current session"""
        ...

    @property
    def extended_metadata(self) -> list[dict[int, Metadata]]:
        """The extended :class:`Metadata` for the current session"""
        ...

    @property
    def calibrations(self) -> dict[int, SensorCalibration]:
        """
        Returns a dict with a :class:`SensorCalibration` per used
        sensor for the current session:

        For example, if session_setup was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 3: SensorConfig()}),
            )

        this attribute will return {1: SensorCalibration(...), 3: SensorCalibration(...)}
        """
        ...

    @property
    def calibrations_provided(self) -> dict[int, bool]:
        """
        Returns whether a calibration was provided for each sensor in
        setup_session. For example, if setup_session was called with

        .. code-block:: python

            client.setup_session(
                SessionConfig({1: SensorConfig(), 2: SensorConfig()}),
                calibrations={2: SensorCalibration(...)},
            )

        this attribute will return ``{1: False, 2: True}``
        """
        ...

    @property
    def _rate_stats(self) -> _RateStats:
        """Returns the data rate statistict from the client"""
        ...



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._utils import get_distances_m
from acconeer.exptool.a121.algo.phase_tracking import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    get_sensor_config,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    sensor_config = get_sensor_config()

    client = a121.Client(**a121.get_client_args(args))
    client.connect()
    metadata = client.setup_session(sensor_config)

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=ProcessorConfig(),
        context=ProcessorContext(),
    )
    pg_updater = PGUpdater(sensor_config, metadata)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()
    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        plot_data = processor.process(result)
        try:
            pg_process.put_data(plot_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(self, sensor_config: a121.SensorConfig, metadata: a121.Metadata):
        (self.distances_m, _) = get_distances_m(sensor_config, metadata)

    def setup(self, win):
        pens = [et.utils.pg_pen_cycler(i) for i in range(3)]
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kws = [dict(pen=pen, **symbol_kw) for pen in pens]

        # sweep and threshold plot
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.setLabel("bottom", "Distance (m)")
        self.sweep_plot.addItem(pg.PlotDataItem())
        self.sweeps_curve = [self.sweep_plot.plot(**feat_kw) for feat_kw in feat_kws]
        self.sweep_vertical_line = pg.InfiniteLine(pen=pens[2])
        self.sweep_plot.addItem(self.sweep_vertical_line)
        self.sweep_smooth_max = et.utils.SmoothMax()
        sweep_plot_legend = pg.LegendItem(offset=(0.0, 0.5))
        sweep_plot_legend.setParentItem(self.sweep_plot)
        sweep_plot_legend.addItem(self.sweeps_curve[0], "Sweep")
        sweep_plot_legend.addItem(self.sweeps_curve[1], "Threshold")

        # argument plot
        argument_plot = win.addPlot(row=1, col=0)
        argument_plot.setMenuEnabled(False)
        argument_plot.showGrid(x=True, y=True)
        argument_plot.addLegend()
        argument_plot.setLabel("bottom", "Distance (m)")
        argument_plot.setLabel("left", "Phase")
        argument_plot.setYRange(-np.pi, np.pi)
        argument_plot.getAxis("left").setTicks(et.utils.pg_phase_ticks)
        argument_plot.setYRange(-np.pi, np.pi)
        argument_plot.addItem(pg.ScatterPlotItem())
        self.argument_curve = argument_plot.plot(
            **dict(pen=None, symbol="o", symbolSize=5, symbolPen="k")
        )
        self.argument_vertical_line = pg.InfiniteLine(pen=pens[2])
        argument_plot.addItem(self.argument_vertical_line)

        # history plot
        self.history_plot = win.addPlot(row=2, col=0)
        self.history_plot.setMenuEnabled(False)
        self.history_plot.showGrid(x=True, y=True)
        self.history_plot.addLegend()
        self.history_plot.setLabel("left", "Distance (mm)")
        self.history_plot.setLabel("bottom", "Time (s)")
        self.history_plot.addItem(pg.PlotDataItem())
        self.history_curve = self.history_plot.plot(**feat_kws[0])

        self.sweep_smooth_max = et.utils.SmoothMax()
        self.distance_hist_smooth_lim = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

    def update(self, processor_result):
        assert processor_result is not None
        assert processor_result.threshold is not None

        sweep = processor_result.lp_abs_sweep
        threshold = processor_result.threshold * np.ones(sweep.size)
        angle_sweep = processor_result.angle_sweep
        peak_loc = processor_result.peak_loc_m
        history = processor_result.distance_history
        rel_time_stamps = processor_result.rel_time_stamps

        # update sweep plot
        self.sweeps_curve[0].setData(self.distances_m, sweep)
        self.sweeps_curve[1].setData(self.distances_m, threshold)
        max_val_in_sweep_plot = max(np.max(sweep), np.max(threshold))
        self.sweep_plot.setYRange(0, self.sweep_smooth_max.update(max_val_in_sweep_plot))

        # update argument plot
        self.argument_curve.setData(self.distances_m, angle_sweep)

        if peak_loc is not None:
            # update vertical lines
            self.sweep_vertical_line.setValue(peak_loc)
            self.argument_vertical_line.setValue(peak_loc)
            self.sweep_vertical_line.show()
            self.argument_vertical_line.show()
        else:
            self.sweep_vertical_line.hide()
            self.argument_vertical_line.hide()

        if history is not None:
            # update history plot
            self.history_curve.setData(rel_time_stamps, history)
            lims = self.distance_hist_smooth_lim.update(history)
            self.history_plot.setYRange(lims[0], lims[1])
        self.history_plot.setXRange(-Processor.TIME_HORIZON_S, 0)


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved
#
from .erroneus_message import ErroneousMessage
from .log_message import LogMessage
from .parse_error import ParseError
from .result_message import EmptyResultMessage, ResultMessage
from .sensor_info_response import SensorInfoResponse
from .set_baudrate_response import SetBaudrateResponse
from .setup_response import SetupResponse
from .streaming_responses import StartStreamingResponse, StopStreamingResponse
from .system_info_response import SystemInfoResponse



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

from enum import Enum, auto
from typing import Callable, Type

import numpy as np

from PySide6.QtWidgets import QWidget

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._plugins import (
    ProcessorBackendPluginBase,
    ProcessorPlotPluginBase,
    ProcessorPluginPreset,
    ProcessorViewPluginBase,
)
from acconeer.exptool.a121.algo.touchless_button import (
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorResult,
    get_close_and_far_processor_config,
    get_close_and_far_sensor_config,
    get_close_processor_config,
    get_close_sensor_config,
)
from acconeer.exptool.app.new import (
    AppModel,
    Message,
    PidgetFactoryMapping,
    PluginFamily,
    PluginGeneration,
    PluginPresetBase,
    PluginSpecBase,
    pidgets,
)


class PluginPresetId(Enum):
    CLOSE_RANGE = auto()
    CLOSE_AND_FAR_RANGE = auto()


class BackendPlugin(ProcessorBackendPluginBase[ProcessorConfig, ProcessorResult]):

    PLUGIN_PRESETS = {
        PluginPresetId.CLOSE_RANGE.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_close_sensor_config()),
            processor_config=get_close_processor_config(),
        ),
        PluginPresetId.CLOSE_AND_FAR_RANGE.value: lambda: ProcessorPluginPreset(
            session_config=a121.SessionConfig(get_close_and_far_sensor_config()),
            processor_config=get_close_and_far_processor_config(),
        ),
    }

    @classmethod
    def get_processor_cls(cls) -> Type[Processor]:
        return Processor

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def get_default_sensor_config(cls) -> a121.SensorConfig:
        return get_close_sensor_config()


class ViewPlugin(ProcessorViewPluginBase[ProcessorConfig]):
    @classmethod
    def get_pidget_mapping(cls) -> PidgetFactoryMapping:
        # Note: Incomplete mapping
        return {
            "sensitivity_close": pidgets.FloatParameterWidgetFactory(
                name_label_text="Sensitivity",
                decimals=1,
                limits=(0.1, 4),
            ),
            "patience_close": pidgets.IntParameterWidgetFactory(
                name_label_text="Patience",
                limits=(0, None),
            ),
            "calibration_duration_s": pidgets.FloatParameterWidgetFactory(
                name_label_text="Calibration duration",
                suffix="s",
                limits=(0, None),
            ),
            "calibration_interval_s": pidgets.FloatParameterWidgetFactory(
                name_label_text="Calibration interval",
                suffix="s",
                limits=(1, None),
            ),
            "measurement_type": pidgets.EnumParameterWidgetFactory(
                enum_type=MeasurementType,
                name_label_text="Range:",
                label_mapping={
                    MeasurementType.CLOSE_RANGE: "Close",
                    MeasurementType.FAR_RANGE: "Far",
                    MeasurementType.CLOSE_AND_FAR_RANGE: "Close and far",
                },
            ),
        }

    @classmethod
    def get_processor_config_cls(cls) -> Type[ProcessorConfig]:
        return ProcessorConfig

    @classmethod
    def supports_multiple_subsweeps(self) -> bool:
        return True


class PlotPlugin(ProcessorPlotPluginBase[ProcessorResult]):
    def __init__(self, *, plot_layout: pg.GraphicsLayout, app_model: AppModel) -> None:
        super().__init__(plot_layout=plot_layout, app_model=app_model)

    def setup(self, metadata: a121.Metadata, sensor_config: a121.SensorConfig) -> None:
        self.detection_history_plot = self._create_detection_plot(self.plot_layout)

        self.detection_history_curve_close = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(1, width=5)
        )
        self.detection_history_curve_far = self.detection_history_plot.plot(
            pen=et.utils.pg_pen_cycler(0, width=5)
        )

        close_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Close detection")
        )
        far_html = (
            '<div style="text-align: center">'
            '<span style="color: #FFFFFF;font-size:15pt;">'
            "{}</span></div>".format("Far detection")
        )
        self.close_text_item = pg.TextItem(
            html=close_html,
            fill=pg.mkColor(0xFF, 0x7F, 0x0E),
            anchor=(0.5, 0),
        )
        self.far_text_item = pg.TextItem(
            html=far_html,
            fill=pg.mkColor(0x1F, 0x77, 0xB4),
            anchor=(0.5, 0),
        )
        pos_left = (100 / 3, 1.8)
        pos_right = (2 * 100 / 3, 1.8)
        self.close_text_item.setPos(*pos_left)
        self.far_text_item.setPos(*pos_right)
        self.detection_history_plot.addItem(self.close_text_item)
        self.detection_history_plot.addItem(self.far_text_item)
        self.close_text_item.hide()
        self.far_text_item.hide()

        self.detection_history = np.full((2, 100), np.NaN)

    def update(self, processor_result: ProcessorResult) -> None:
        detection = np.array([processor_result.detection_close, processor_result.detection_far])
        self.detection_history = np.roll(self.detection_history, -1, axis=1)
        self.detection_history[:, -1] = detection

        self.detection_history_curve_close.setData(self.detection_history[0])
        self.detection_history_curve_far.setData(self.detection_history[1])

        if processor_result.detection_close:
            self.close_text_item.show()
        else:
            self.close_text_item.hide()

        if processor_result.detection_far:
            self.far_text_item.show()
        else:
            self.far_text_item.hide()

    @staticmethod
    def _create_detection_plot(parent: pg.GraphicsLayout) -> pg.PlotItem:
        detection_history_plot = parent.addPlot()
        detection_history_plot.setMenuEnabled(False)
        detection_history_plot.setMouseEnabled(x=False, y=False)
        detection_history_plot.hideButtons()
        detection_history_plot.showGrid(x=True, y=True, alpha=0.5)
        detection_history_plot.setYRange(-0.1, 1.8)
        return detection_history_plot


class PluginSpec(PluginSpecBase):
    def create_backend_plugin(
        self, callback: Callable[[Message], None], key: str
    ) -> BackendPlugin:
        return BackendPlugin(callback=callback, generation=self.generation, key=key)

    def create_view_plugin(self, app_model: AppModel, view_widget: QWidget) -> ViewPlugin:
        return ViewPlugin(app_model=app_model, view_widget=view_widget)

    def create_plot_plugin(
        self, app_model: AppModel, plot_layout: pg.GraphicsLayout
    ) -> PlotPlugin:
        return PlotPlugin(app_model=app_model, plot_layout=plot_layout)


TOUCHLESS_BUTTON_PLUGIN = PluginSpec(
    generation=PluginGeneration.A121,
    key="touchless_button",
    title="Touchless button",
    description="Detect tap/wave motion and register as button press.",
    family=PluginFamily.EXAMPLE_APP,
    presets=[
        PluginPresetBase(
            name="Close range",
            description="Close range",
            preset_id=PluginPresetId.CLOSE_RANGE,
        ),
        PluginPresetBase(
            name="Close and far range",
            description="Close and far range",
            preset_id=PluginPresetId.CLOSE_AND_FAR_RANGE,
        ),
    ],
    default_preset_id=PluginPresetId.CLOSE_RANGE,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import importlib.resources
import logging
from enum import Enum
from functools import partial
from typing import Any, Optional

import qtawesome as qta

from PySide6 import QtCore
from PySide6.QtSvgWidgets import QSvgWidget
from PySide6.QtWidgets import (
    QButtonGroup,
    QFrame,
    QGraphicsOpacityEffect,
    QGridLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

import pyqtgraph as pg

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.app.new._enums import PluginFamily
from acconeer.exptool.app.new.app_model import AppModel, PluginPresetSpec, PluginSpec
from acconeer.exptool.app.new.pluginbase import PlotPluginBase, PluginSpecBase

from .plugin_components.utils import VerticalGroupBox
from .utils import HorizontalSeparator, ScrollAreaDecorator, TopAlignDecorator


log = logging.getLogger(__name__)


class PluginSelectionButton(QPushButton):
    plugin: PluginSpec

    def __init__(self, plugin: PluginSpecBase, parent: QWidget) -> None:
        super().__init__(parent)

        self.plugin = plugin

        self.setText(plugin.title)
        self.setStyleSheet("text-align: left; font-weight: bold;")
        self.setCheckable(True)


class PluginSelectionButtonGroup(QButtonGroup):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setExclusive(True)

    def addButton(self, button: PluginSelectionButton) -> None:
        super().addButton(button)

    def checkedButton(self) -> PluginSelectionButton:
        button = super().checkedButton()
        assert isinstance(button, PluginSelectionButton)
        return button

    def buttons(self) -> list[PluginSelectionButton]:
        buttons = super().buttons()
        assert isinstance(buttons, list)
        assert all(isinstance(e, PluginSelectionButton) for e in buttons)
        return buttons


class PluginPresetButton(QPushButton):
    def __init__(self, plugin_preset: PluginPresetSpec, default: bool, parent: QWidget) -> None:
        super().__init__(parent)
        if default:
            self.setText(f"{plugin_preset.name} (default)")
        else:
            self.setText(plugin_preset.name)
        if plugin_preset.description is not None:
            self.setToolTip(plugin_preset.description)
        self.setStyleSheet("text-align: left; font-weight: bold;")


class PluginPresetPlaceholder(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        app_model.sig_load_plugin.connect(self._on_load_plugin)

        self.app_model = app_model
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)
        self.group_box = VerticalGroupBox("Preset Configurations", self)
        self.layout().addWidget(self.group_box)
        self.preset_buttons_widget: Optional[QWidget] = None

    def _clear(self) -> None:
        if self.preset_buttons_widget is not None:
            self.preset_buttons_widget.deleteLater()
            self.preset_buttons_widget = None

    def _on_preset_click(self, preset_id: Enum) -> None:
        self.app_model.set_plugin_preset(preset_id)

    def _on_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        self._clear()
        if plugin_spec is not None and plugin_spec.presets is not None:
            self.preset_buttons_widget = QWidget(self.group_box)
            self.preset_buttons_widget.setLayout(QVBoxLayout(self.preset_buttons_widget))
            self.preset_buttons_widget.layout().setContentsMargins(0, 0, 0, 0)
            self.group_box.layout().addWidget(self.preset_buttons_widget)

            for preset in plugin_spec.presets:
                default_preset = (
                    len(plugin_spec.presets) > 1
                    and preset.preset_id == plugin_spec.default_preset_id
                    and preset.name.lower() != "default"
                )
                button = PluginPresetButton(preset, default_preset, self)
                self.preset_buttons_widget.layout().addWidget(button)
                button.clicked.connect(partial(self._on_preset_click, preset.preset_id))


class PluginSelection(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        app_model.sig_notify.connect(self._on_app_model_update)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(11)

        group_boxes = {}
        for family in PluginFamily:
            group_box = QGroupBox(self)
            group_box.setTitle(family.value)
            group_box.setHidden(True)
            group_box.setLayout(QVBoxLayout(group_box))
            self.layout().addWidget(group_box)
            group_boxes[family] = group_box

        self.button_group = PluginSelectionButtonGroup(self)
        self.button_group.buttonClicked.connect(self._on_load_click)

        for plugin in app_model.plugins:
            assert isinstance(plugin, PluginSpecBase)
            group_box = group_boxes[plugin.family]
            group_box.setHidden(False)

            button = PluginSelectionButton(plugin, group_box)
            self.button_group.addButton(button)
            group_box.layout().addWidget(button)

            if plugin.description:
                label = QLabel(group_box)
                label.setText(plugin.description)
                label.setWordWrap(True)
                group_box.layout().addWidget(label)

    def _on_load_click(self) -> None:
        self.layout().addStretch(1)

        plugin = self.button_group.checkedButton().plugin
        self.app_model.load_plugin(plugin)

    def _on_app_model_update(self, app_model: AppModel) -> None:
        plugin: Optional[PluginSpec] = app_model.plugin

        if plugin is None:
            self.button_group.setExclusive(False)

            for button in self.button_group.buttons():
                button.setChecked(False)

            self.button_group.setExclusive(True)
        else:
            buttons = self.button_group.buttons()
            button = next(b for b in buttons if b.plugin == plugin)
            button.setChecked(True)

        self.setEnabled(app_model.plugin_state.is_steady)


class PluginSelectionArea(QWidget):
    def __init__(self, app_model: AppModel, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)
        self.setLayout(QVBoxLayout(self))
        self.layout().addWidget(
            ScrollAreaDecorator(TopAlignDecorator(PluginSelection(app_model, self)))
        )
        self.layout().addWidget(HorizontalSeparator())
        self.layout().addWidget(PluginPresetPlaceholder(app_model, self))


class PlotPlaceholder(QWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().addStretch(1)
        self.layout().addLayout(self._select_module_text())
        self.layout().addWidget(self._teaser_text())
        self.layout().addStretch(1)

    def _select_module_text(self) -> QHBoxLayout:
        h_box = QHBoxLayout()
        h_box.addStretch(1)

        icon_widget = qta.IconWidget()
        icon_widget.setIconSize(QtCore.QSize(36, 36))
        icon_widget.setIcon(qta.icon("ph.arrow-left-bold", color="#4d5157"))

        label = QLabel("Select a module to begin", self)
        label.setStyleSheet("font-size: 20px;")
        label.setAlignment(QtCore.Qt.AlignCenter)

        h_box.addWidget(icon_widget)
        h_box.addWidget(label)
        h_box.addStretch(1)

        return h_box

    def _teaser_text(self) -> QLabel:
        teaser_label = QLabel(
            "More detectors and example applications will\nbe released continuously.", self
        )
        teaser_label.setStyleSheet("font-size: 16px;")
        teaser_label.setAlignment(QtCore.Qt.AlignCenter)

        return teaser_label


class PluginPlotArea(QFrame):
    _FPS = 60

    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.child_widget = PlotPlaceholder()
        self.plot_plugin: Optional[PlotPluginBase] = None

        self.setObjectName("PluginPlotArea")
        self.setStyleSheet("QFrame#PluginPlotArea {background: #fff; border: 0;}")
        self.setFrameStyle(0)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(self.child_widget)

        self.startTimer(int(1000 / self._FPS))

        app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)
        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            raise RuntimeError(f"{type(app_model.plugin)} is not a PluginSpecBase.")

    def timerEvent(self, event: QtCore.QTimerEvent) -> None:
        if self.plot_plugin is None:
            return

        self.plot_plugin.draw()

    def _on_app_model_load_plugin(self, plugin: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its plot_plugin "
            + f"({self.plot_plugin.__class__.__name__}) in favour of {plugin}"
        )
        self.plot_plugin = None
        self.child_widget.deleteLater()

        if plugin is None:
            self.child_widget = PlotPlaceholder()
        else:
            self.child_widget = pg.GraphicsLayoutWidget()
            self.plot_plugin = plugin.create_plot_plugin(
                app_model=self.app_model,
                plot_layout=self.child_widget.ci,
            )

        self.layout().addWidget(self.child_widget)


class ControlPlaceholder(QWidget):
    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent)

        self.setLayout(QGridLayout(self))

        with importlib.resources.path(resources, "icon-black.svg") as path:
            icon = QSvgWidget(str(path), self)

        icon.setMaximumSize(60, 60)
        icon.renderer().setAspectRatioMode(QtCore.Qt.KeepAspectRatio)
        effect = QGraphicsOpacityEffect(icon)
        effect.setOpacity(0.1)
        icon.setGraphicsEffect(effect)

        self.layout().addWidget(icon)


class PluginControlArea(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model
        self.app_model.sig_load_plugin.connect(self._on_app_model_load_plugin)

        self.child_widget: QWidget = ControlPlaceholder()
        self.view_plugin: Optional[Any] = None

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.layout().addWidget(self.child_widget)

        if isinstance(app_model.plugin, PluginSpecBase):
            self._on_app_model_load_plugin(app_model.plugin)
        elif app_model.plugin is not None:
            raise RuntimeError(f"{type(app_model.plugin)} is not a PluginSpecBase.")

    def _on_app_model_load_plugin(self, plugin_spec: Optional[PluginSpecBase]) -> None:
        log.debug(
            f"{self.__class__.__name__} is going to replace its view_plugin"
            + f"({self.view_plugin.__class__.__name__}) in favour of {plugin_spec}"
        )
        self.view_plugin = None
        self.child_widget.deleteLater()

        if plugin_spec is None:
            self.child_widget = ControlPlaceholder()
        else:
            self.child_widget = QWidget()
            self.view_plugin = plugin_spec.create_view_plugin(
                app_model=self.app_model,
                view_widget=self.child_widget,
            )

        self.layout().addWidget(self.child_widget)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._factory import get_exploration_protocol
from ._latest import ExplorationProtocol, ExplorationProtocolError
from ._no_5_2_mhz import ExplorationProtocol_No_5_2MHz_PRF
from ._no_15_6_mhz import ExplorationProtocol_No_15_6MHz_PRF
from ._no_calibration_reuse import ExplorationProtocol_NoCalibrationReuse
from .server_error import ServerError



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import List, Optional

import attrs

import acconeer.exptool as et
from acconeer.exptool.a121._core.entities import ClientInfo, SensorCalibration, SessionConfig
from acconeer.exptool.a121._core.mediators import BufferedLink, ClientError
from acconeer.exptool.a121._core.utils import iterate_extended_structure
from acconeer.exptool.utils import SerialDevice, USBDevice  # type: ignore[import]

from .links import AdaptedSerialLink, AdaptedSocketLink, AdaptedUSBLink, NullLink


def get_one_serial_device() -> SerialDevice:
    acconeer_serial_devices: List[SerialDevice] = [
        device for device in et.utils.get_serial_devices() if device.name is not None
    ]

    if not acconeer_serial_devices:
        raise ClientError("No serial devices detected. Cannot auto detect.")
    elif len(acconeer_serial_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in acconeer_serial_devices])
        raise ClientError("There are multiple devices detected. Specify one:\n" + devices_string)
    else:
        return acconeer_serial_devices[0]


def determine_serial_device(serial_port: Optional[str]) -> str:
    if serial_port is None:
        return str(get_one_serial_device().port)
    else:
        return serial_port


def get_one_usb_device(only_accessible: bool = False) -> USBDevice:
    usb_devices = et.utils.get_usb_devices(only_accessible=only_accessible)
    if not usb_devices:
        raise ClientError("No USB devices detected. Cannot auto detect.")
    elif len(usb_devices) > 1:
        devices_string = "".join([f" - {dev}\n" for dev in usb_devices])
        raise ClientError("There are multiple devices detected. Specify one:\n" + devices_string)
    else:
        return usb_devices[0]


def determine_usb_device(usb_device: Optional[USBDevice]) -> USBDevice:
    if usb_device is None:
        return get_one_usb_device(only_accessible=True)
    else:
        return usb_device


def link_factory(client_info: ClientInfo) -> BufferedLink:

    if client_info.ip_address is not None:
        return AdaptedSocketLink(host=client_info.ip_address)

    if client_info.serial_port is not None:
        link = AdaptedSerialLink(
            port=client_info.serial_port,
        )

        return link

    if client_info.usb_device is not None:
        link = AdaptedUSBLink(
            vid=client_info.usb_device.vid,
            pid=client_info.usb_device.pid,
            serial=client_info.usb_device.serial,
        )

        return link

    return NullLink()


def autodetermine_client_link(client_info: ClientInfo) -> ClientInfo:
    error_message = ""
    try:
        client_info = attrs.evolve(
            client_info,
            usb_device=determine_usb_device(client_info.usb_device),
        )

        return client_info
    except ClientError as exc:
        error_message += f"\nUSB: {str(exc)}"
        pass

    try:
        client_info = attrs.evolve(
            client_info,
            serial_port=determine_serial_device(client_info.serial_port),
        )
        return client_info
    except ClientError as exc:
        error_message += f"\nSerial: {str(exc)}"

    raise ClientError(f"Cannot auto detect:{error_message}")


def get_calibrations_provided(
    session_config: SessionConfig,
    calibrations: Optional[dict[int, SensorCalibration]] = None,
) -> dict[int, bool]:
    calibrations_provided = {}
    for _, sensor_id, _ in iterate_extended_structure(session_config.groups):
        if calibrations:
            calibrations_provided[sensor_id] = sensor_id in calibrations
        else:
            calibrations_provided[sensor_id] = False

    return calibrations_provided



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .client import Client
from .exploration_client import ExplorationClient
from .exploration_protocol import (
    ExplorationProtocol,
    ExplorationProtocolError,
    ServerError,
    get_exploration_protocol,
)
from .links import AdaptedSerialLink, AdaptedSocketLink, AdaptedUSBLink
from .message import Message, MessageT



# Copyright (c) Acconeer AB, 2022
# All rights reserved

SDK_VERSION = "2.14.0"

from . import recording
from ._clients import Client, Link, Protocol
from ._clients.base import SessionSetupError
from ._configs import (
    EnvelopeServiceConfig,
    IQServiceConfig,
    PowerBinServiceConfig,
    SparseServiceConfig,
)
from ._modes import Mode, get_mode
from ._utils import ExampleArgumentParser, get_client_args, get_range_depths



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

# Added here to force pyqtgraph to choose PySide
import PySide6  # noqa: F401

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo._utils import APPROX_BASE_STEP_LENGTH_M
from acconeer.exptool.a121.algo.vibration import (
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorResult,
    get_sensor_config,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    sensor_config = get_sensor_config()

    client = a121.Client(**a121.get_client_args(args))
    client.connect()
    metadata = client.setup_session(sensor_config)

    processor = Processor(
        sensor_config=sensor_config,
        metadata=metadata,
        processor_config=ProcessorConfig(),
        context=ProcessorContext(),
    )
    pg_updater = PGUpdater(sensor_config)
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()
    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        result = client.get_next()
        plot_data = processor.process(result)
        try:
            pg_process.put_data(plot_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(self, sensor_config):
        self.meas_dist_m = sensor_config.start_point * APPROX_BASE_STEP_LENGTH_M

    def setup(self, win):

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        symbol_dot_kw = dict(symbol="o", symbolSize=10, symbolBrush=brush, symbolPen="k")

        # precense plot
        self.precense_plot = pg.PlotItem()
        self.precense_plot.setMenuEnabled(False)
        self.precense_plot.showGrid(x=False, y=True)
        self.precense_plot.setLabel("left", "Max amplitude")
        self.precense_plot.setLabel("bottom", "Distance (m)")
        self.precense_plot.setXRange(self.meas_dist_m - 0.001, self.meas_dist_m + 0.001)
        self.precense_curve = self.precense_plot.plot(**dict(pen=pen, **symbol_dot_kw))

        self.smooth_max_precense = et.utils.SmoothMax(tau_decay=10.0)

        # sweep and threshold plot
        self.time_series_plot = pg.PlotItem()
        self.time_series_plot.setMenuEnabled(False)
        self.time_series_plot.showGrid(x=True, y=True)
        self.time_series_plot.setLabel("left", "Displacement (mm)")
        self.time_series_plot.setLabel("bottom", "History")
        self.time_series_curve = self.time_series_plot.plot(**feat_kw)

        sublayout = win.addLayout(row=0, col=0)
        sublayout.layout.setColumnStretchFactor(1, 5)
        sublayout.addItem(self.precense_plot, row=0, col=0)
        sublayout.addItem(self.time_series_plot, row=0, col=1)

        self.smooth_lim_time_series = et.utils.SmoothLimits(tau_decay=0.5, tau_grow=0.1)

        self.fft_plot = win.addPlot(col=0, row=1)
        self.fft_plot.setMenuEnabled(False)
        self.fft_plot.showGrid(x=True, y=True)
        self.fft_plot.setLabel("left", "Power (db)")
        self.fft_plot.setLabel("bottom", "Frequency (Hz)")
        self.fft_plot.addItem(pg.PlotDataItem())
        self.fft_curve = self.fft_plot.plot(**feat_kw)

        self.smooth_max_fft = et.utils.SmoothMax()

    def update(self, processor_result: ProcessorResult) -> None:

        time_series = processor_result.time_series
        z_abs_db = processor_result.lp_z_abs_db
        freqs = processor_result.freqs
        max_amplitude = processor_result.max_amplitude

        self.precense_curve.setData([self.meas_dist_m], [max_amplitude])
        lim = self.smooth_max_precense.update(max_amplitude)
        self.precense_plot.setYRange(0, max(1000.0, lim))

        self.time_series_curve.setData(time_series)
        lim = self.smooth_lim_time_series.update(time_series)
        self.time_series_plot.setYRange(lim[0], lim[1])

        self.fft_curve.setData(freqs, z_abs_db)
        lim = self.smooth_max_fft.update(np.max(z_abs_db))
        self.fft_plot.setYRange(0, lim)


if __name__ == "__main__":
    main()



class WinUSBError(Exception):
    def __init__(self, reason, response=None):
        self.reason = reason
        self.response = response
        Exception.__init__(self, reason)

    def __str__(self):
        return self.reason



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from acconeer.exptool.a111._clients.links import (  # type: ignore[import]
    ExploreSerialLink,
    SocketLink,
    USBLink,
)
from acconeer.exptool.a121._core.mediators import BufferedLink


class NullLinkError(RuntimeError):
    pass


class NullLink(BufferedLink):
    ERROR = NullLinkError("Link is undetermined.")
    """Link null object.

    :raises: ``RuntimeError`` if any of its methods is called.
    """

    def connect(self) -> None:
        raise self.ERROR

    @property
    def timeout(self) -> float:
        raise self.ERROR

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        raise self.ERROR

    def recv(self, num_bytes: int) -> bytes:
        raise self.ERROR

    def send(self, bytes_: bytes) -> None:
        raise self.ERROR

    def disconnect(self) -> None:
        raise self.ERROR

    def recv_until(self, byte_sequence: bytes) -> bytes:
        raise self.ERROR


class AdaptedSocketLink(SocketLink):
    """This subclass only adapts the signature.
    Positional arguments would've executed fine.
    """

    def recv_until(self, byte_sequence: bytes) -> bytes:
        return bytes(super().recv_until(bs=byte_sequence))

    def send(self, bytes_: bytes) -> None:
        super().send(data=bytes_)


class AdaptedSerialLink(ExploreSerialLink):
    """This subclass only adapts the signature.
    Positional arguments would've executed fine.
    """

    def recv_until(self, byte_sequence: bytes) -> bytes:
        return bytes(super().recv_until(bs=byte_sequence))

    def send(self, bytes_: bytes) -> None:
        super().send(data=bytes_)


class AdaptedUSBLink(USBLink):
    """This subclass only adapts the signature.
    Positional arguments would've executed fine.
    """

    def recv_until(self, byte_sequence: bytes) -> bytes:
        return bytes(super().recv_until(bs=byte_sequence))

    def send(self, bytes_: bytes) -> None:
        super().send(data=bytes_)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import enum
import os
import sys
from argparse import SUPPRESS, ArgumentParser
from pathlib import Path
from typing import Optional

from PySide6.QtWidgets import QCheckBox, QLabel, QLineEdit, QPushButton


class LoadState(enum.Enum):
    UNLOADED = enum.auto()
    BUFFERED = enum.auto()
    LOADED = enum.auto()


class ErrorFormater:
    def __init__(self):
        pass

    def error_to_text(self, error):
        exc_type, exc_obj, exc_tb = sys.exc_info()
        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
        err_text = "File: {}<br>Line: {}<br>Error: {}".format(fname, exc_tb.tb_lineno, error)

        return err_text


class Count:
    def __init__(self, val=0):
        self.val = val

    def pre_incr(self):
        self.val += 1
        return self.val

    def post_incr(self):
        ret = self.val
        self.val += 1
        return ret

    def decr(self, val=1):
        self.val -= val

    def set_val(self, val):
        self.val = val


class ExptoolArgumentParser(ArgumentParser):
    def __init__(self):
        super().__init__()
        self.add_argument(
            "--purge-config",
            action="store_true",
            help="Remove Exptool-related files interactively.",
        )
        self.add_argument(
            "--no-config",
            action="store_false",
            dest="use_last_config",
            help="Runs Exptool without loading or saving gui configuration.",
        )
        self.add_argument(
            "--portable",
            action="store_true",
            help=SUPPRESS,  # makes option hidden
        )


class CalibrationStatus(enum.Enum):
    NONE = enum.auto()
    BUFFERED = enum.auto()
    IN_PROCESSOR = enum.auto()


class CalibrationUiState:
    """
    Presenter object that modifies the semi-passive view that is
    the calibration-related ui-elements
    """

    def __init__(
        self,
        load_btn: Optional[QPushButton] = None,
        save_btn: Optional[QPushButton] = None,
        clear_btn: Optional[QPushButton] = None,
        source_text: Optional[QLineEdit] = None,
        status_label: Optional[QLabel] = None,
        auto_apply_cb: Optional[QCheckBox] = None,
        apply_btn: Optional[QPushButton] = None,
    ):
        self._source: Optional[str] = None
        self._modified: bool = False
        self._auto_apply: bool = False
        self._scan_is_running: bool = False  # Will be updated from GUI
        self._calibration_status: CalibrationStatus = CalibrationStatus.NONE

        self._load_btn = load_btn
        self._save_btn = save_btn
        self._clear_btn = clear_btn
        self._source_text = source_text
        self._status_label = status_label
        self._auto_apply_cb = auto_apply_cb
        self._apply_btn = apply_btn
        self._update_ui_elements()

    def get_status_tooltip_text(self):
        return "<br>".join(
            [
                "<i>None</i>:",
                "No calibration has been made.",
                "",
                "<i>Buffered</i>:",
                "A calibration has been made but it is not used.",
                "",
                "<i>Used in processor</i>:",
                "A calibration is/will be used.",
            ]
        )

    def get_displayed_status_text(self):
        TEXT_MAP = {
            CalibrationStatus.NONE: "None",
            CalibrationStatus.BUFFERED: "Buffered",
            CalibrationStatus.IN_PROCESSOR: "Used in processor",
        }

        return f"<i>{TEXT_MAP[self.calibration_status]}</i>"

    def get_source_tooltip_text(self):
        """Produces a nice tooltip from the `self.source` field"""
        if self.source is None:
            return "No calibration loaded. Load from file or start a measurement to get one."

        if self.source == "Session":
            return "Calibration was created in this session and is not saved to a file."
        else:
            filename = self.source
            return f"Calibration is loaded from {filename}"

    def get_displayed_source_text(self):
        """
        What should be displayed in the "Calibration source" field. Should be one of
            1. <file_name>[*]
            2. Session[*]
            3. <empty string>, which makes the placeholder text show in case of a QLineEdit.
        Where the asterisk denotes if there are unsaved changes.
        """
        if self.source is None:
            return ""

        source_str = self.source

        if source_str != "Session":
            path = Path(source_str)

            file_str = path.name
            if path.parent != Path("."):
                # Truncated dir path. Whole path will be shown in tooltip.
                file_str = ".../" + file_str
            source_str = file_str

        mby_asterisk = "*" if self.modified else ""
        return f"{source_str}{mby_asterisk}"

    def set_scan_is_running(self, value: bool):
        self._scan_is_running = value
        self._update_ui_elements()

    def clear(self):
        self.source = None
        self.modified = False
        self.calibration_status = CalibrationStatus.NONE

    def save(self, save_destination):
        self.source = save_destination
        self.modified = False

    def load(self, source):
        self.source = source
        self.modified = False
        self.calibration_status = CalibrationStatus.BUFFERED

    def buffer(self, source):
        self.source = source
        self.modified = True
        self.calibration_status = CalibrationStatus.BUFFERED

    def _update_ui_elements(self):
        if self._load_btn:
            self._load_btn.setEnabled(self.load_button_enabled)

        if self._save_btn:
            self._save_btn.setEnabled(self.save_button_enabled)

        if self._clear_btn:
            self._clear_btn.setEnabled(self.clear_button_enabled)

        if self._source_text:
            self._source_text.setText(self.get_displayed_source_text())
            self._source_text.setToolTip(self.get_source_tooltip_text())
            if self.is_display_text_italic:
                self._source_text.setStyleSheet("QLineEdit { font: italic }")
            else:
                self._source_text.setStyleSheet("")

        if self._status_label:
            self._status_label.setText(self.get_displayed_status_text())
            self._status_label.setToolTip(self.get_status_tooltip_text())
        if self._auto_apply_cb:
            self._auto_apply_cb.setChecked(self._auto_apply)

        if self._apply_btn:
            self._apply_btn.setEnabled(self.apply_button_enabled)

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, value):
        self._source = value
        self._update_ui_elements()

    @property
    def modified(self):
        return self._modified

    @modified.setter
    def modified(self, value):
        self._modified = value
        self._update_ui_elements()

    @property
    def calibration_status(self):
        return self._calibration_status

    @calibration_status.setter
    def calibration_status(self, value):
        self._calibration_status = value
        self._update_ui_elements()

    @property
    def is_display_text_italic(self):
        """It's pretty nice if the text is in italics if there are unsaved changes."""
        return self.modified

    @property
    def load_button_enabled(self):
        """
        The load button should be enabled if the current calibration is cleared or saved
        I.e. calibration has no unsaved changes state.
        """
        return not self.modified

    @property
    def save_button_enabled(self):
        """The save button should be enabled if there are unsaved changes."""
        return self.modified

    @property
    def _has_source(self):
        return self.source is not None

    clear_button_enabled = _has_source

    @property
    def apply_button_enabled(self):
        """The apply button should only work in a "LIVE" scenario."""
        return (
            self._has_source
            and self._scan_is_running
            and self.calibration_status is not CalibrationStatus.IN_PROCESSOR
        )

    @property
    def auto_apply(self):
        return self._auto_apply

    @auto_apply.setter
    def auto_apply(self, auto_apply):
        self._auto_apply = auto_apply
        self._update_ui_elements()



from ctypes import *
from ctypes.wintypes import *

from .winusbclasses import (
    GUID,
    Overlapped,
    PipeInfo,
    SpDeviceInterfaceData,
    SpDeviceInterfaceDetailData,
    SpDevinfoData,
    UsbInterfaceDescriptor,
)


WinUsb_Initialize = "WinUsb_Initialize"
WinUsb_ControlTransfer = "WinUsb_ControlTransfer"
WinUsb_GetDescriptor = "WinUsb_GetDescriptor"
WinUsb_GetOverlappedResult = "WinUsb_GetOverlappedResult"
WinUsb_SetPipePolicy = "WinUsb_SetPipePolicy"
WinUsb_ReadPipe = "WinUsb_ReadPipe"
WinUsb_WritePipe = "WinUsb_WritePipe"
WinUsb_FlushPipe = "WinUsb_FlushPipe"
WinUsb_Free = "WinUsb_Free"
WinUsb_QueryDeviceInformation = "WinUsb_QueryDeviceInformation"
WinUsb_QueryInterfaceSettings = "WinUsb_QueryInterfaceSettings"
WinUsb_GetAssociatedInterface = "WinUsb_GetAssociatedInterface"
WinUsb_QueryPipe = "WinUsb_QueryPipe"
# WinUsb_ControlTransfer = "WinUsb_ControlTransfer"
# WinUsb_QueryPipe = "WinUsb_QueryPipe"
Close_Handle = "CloseHandle"
CreateFile = "CreateFileW"
ReadFile = "ReadFile"
CancelIo = "CancelIo"
WriteFile = "WriteFile"
SetEvent = "SetEvent"
WaitForSingleObject = "WaitForSingleObject"
GetLastError = "GetLastError"
SetupDiGetClassDevs = "SetupDiGetClassDevs"
SetupDiEnumDeviceInterfaces = "SetupDiEnumDeviceInterfaces"
SetupDiGetDeviceInterfaceDetail = "SetupDiGetDeviceInterfaceDetail"
SetupDiGetDeviceRegistryProperty = "SetupDiGetDeviceRegistryProperty"
SetupDiEnumDeviceInfo = "SetupDiEnumDeviceInfo"

SPDRP_HARDWAREID = 1
SPDRP_FRIENDLYNAME = 12
SPDRP_LOCATION_PATHS = 35
SPDRP_MFG = 11


def get_winusb_functions(windll):
    """Functions availabe from WinUsb dll and their types"""
    winusb_dict = {}
    winusb_functions = {}
    winusb_restypes = {}
    winusb_argtypes = {}

    # BOOL __stdcall WinUsb_Initialize( _In_ HANDLE DeviceHandle,_Out_  PWINUSB_INTERFACE_HANDLE InterfaceHandle);
    winusb_functions[WinUsb_Initialize] = windll.WinUsb_Initialize
    winusb_restypes[WinUsb_Initialize] = BOOL
    winusb_argtypes[WinUsb_Initialize] = [HANDLE, POINTER(c_void_p)]

    # BOOL __stdcall WinUsb_ControlTransfer(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ WINUSB_SETUP_PACKET SetupPacket, _Out_ PUCHAR Buffer,_In_ ULONG BufferLength,_Out_opt_  PULONG LengthTransferred,_In_opt_  LPOVERLAPPED Overlapped);
    winusb_functions[WinUsb_ControlTransfer] = windll.WinUsb_ControlTransfer
    # winusb_restypes[WinUsb_ControlTransfer] = BOOL
    # winusb_argtypes[WinUsb_ControlTransfer] = [c_void_p, UsbSetupPacket, POINTER(c_ubyte), c_ulong, POINTER(c_ulong), LpOverlapped]

    # BOOL __stdcall WinUsb_GetDescriptor(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ UCHAR DescriptorType,_In_ UCHAR Index,_In_ USHORT LanguageID,_Out_ PUCHAR Buffer,_In_ ULONG BufferLength,_Out_ PULONG LengthTransferred);
    winusb_functions[WinUsb_GetDescriptor] = windll.WinUsb_GetDescriptor
    winusb_restypes[WinUsb_GetDescriptor] = BOOL
    winusb_argtypes[WinUsb_GetDescriptor] = [
        c_void_p,
        c_ubyte,
        c_ubyte,
        c_ushort,
        POINTER(c_ubyte),
        c_ulong,
        POINTER(c_ulong),
    ]

    # BOOL __stdcall WinUsb_GetOverlappedResult(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ LPOVERLAPPED lpOverlapped,_Out_ LPDWORD lpNumberOfBytesTransferred,_In_ BOOL bWait);
    winusb_functions[WinUsb_GetOverlappedResult] = windll.WinUsb_GetOverlappedResult
    winusb_restypes[WinUsb_GetOverlappedResult] = BOOL
    # winusb_argtypes[WinUsb_GetOverlappedResult] = []

    # BOOL __stdcall WinUsb_ReadPipe( _In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ UCHAR PipeID,_Out_ PUCHAR Buffer,_In_ ULONG BufferLength,_Out_opt_ PULONG LengthTransferred,_In_opt_ LPOVERLAPPED Overlapped);
    winusb_functions[WinUsb_ReadPipe] = windll.WinUsb_ReadPipe
    # winusb_restypes[WinUsb_ReadPipe] = BOOL
    # winusb_argtypes[WinUsb_ReadPipe] = [c_void_p, c_ubyte, POINTER(c_ubyte), c_ulong, POINTER(c_ulong), LpOverlapped]

    # BOOL __stdcall WinUsb_ReadPipe( _In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ UCHAR PipeID,_Out_ PUCHAR Buffer,_In_ ULONG BufferLength,_Out_opt_ PULONG LengthTransferred,_In_opt_ LPOVERLAPPED Overlapped);
    winusb_functions[WinUsb_SetPipePolicy] = windll.WinUsb_SetPipePolicy
    winusb_restypes[WinUsb_SetPipePolicy] = BOOL
    winusb_argtypes[WinUsb_SetPipePolicy] = [c_void_p, c_ubyte, c_ulong, c_ulong, c_void_p]

    # BOOL __stdcall WinUsb_WritePipe(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ UCHAR PipeID,_In_ PUCHAR Buffer,_In_ ULONG BufferLength,_Out_opt_  PULONG LengthTransferred,_In_opt_ LPOVERLAPPED Overlapped);
    winusb_functions[WinUsb_WritePipe] = windll.WinUsb_WritePipe
    # winusb_restypes[WinUsb_WritePipe] = BOOL
    # winusb_argtypes[WinUsb_WritePipe] = [c_void_p, c_ubyte, POINTER(c_ubyte), c_ulong, POINTER(c_ulong), LpOverlapped]

    # BOOL __stdcall WinUsb_FlushPipe(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle);
    winusb_functions[WinUsb_FlushPipe] = windll.WinUsb_FlushPipe
    winusb_restypes[WinUsb_FlushPipe] = BOOL
    winusb_argtypes[WinUsb_FlushPipe] = [c_void_p, c_ubyte]

    # BOOL __stdcall WinUsb_Free(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle);
    winusb_functions[WinUsb_Free] = windll.WinUsb_Free
    winusb_restypes[WinUsb_Free] = BOOL
    winusb_argtypes[WinUsb_Free] = [c_void_p]

    # BOOL __stdcall WinUsb_QueryDeviceInformation(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ ULONG InformationType,_Inout_ PULONG BufferLength,_Out_ PVOID Buffer);
    winusb_functions[WinUsb_QueryDeviceInformation] = windll.WinUsb_QueryDeviceInformation
    winusb_restypes[WinUsb_QueryDeviceInformation] = BOOL
    winusb_argtypes[WinUsb_QueryDeviceInformation] = [
        c_void_p,
        c_ulong,
        POINTER(c_ulong),
        c_void_p,
    ]

    # BOOL __stdcall WinUsb_QueryInterfaceSettings(_In_ WINUSB_INTERFACE_HANDLE InterfaceHandle,_In_ UCHAR AlternateSettingNumber,_Out_ PUSB_INTERFACE_DESCRIPTOR UsbAltInterfaceDescriptor);
    winusb_functions[WinUsb_QueryInterfaceSettings] = windll.WinUsb_QueryInterfaceSettings
    winusb_restypes[WinUsb_QueryInterfaceSettings] = BOOL
    winusb_argtypes[WinUsb_QueryInterfaceSettings] = [
        c_void_p,
        c_ubyte,
        POINTER(UsbInterfaceDescriptor),
    ]

    winusb_functions[WinUsb_QueryPipe] = windll.WinUsb_QueryPipe
    winusb_restypes[WinUsb_QueryPipe] = BOOL
    winusb_argtypes[WinUsb_QueryPipe] = [c_void_p, c_ubyte, c_ubyte, POINTER(PipeInfo)]

    winusb_functions[WinUsb_GetAssociatedInterface] = windll.WinUsb_GetAssociatedInterface
    winusb_restypes[WinUsb_GetAssociatedInterface] = BOOL
    winusb_argtypes[WinUsb_GetAssociatedInterface] = [c_void_p, c_ubyte, POINTER(c_void_p)]

    winusb_dict["functions"] = winusb_functions
    winusb_dict["restypes"] = winusb_restypes
    winusb_dict["argtypes"] = winusb_argtypes
    return winusb_dict


def get_kernel32_functions(kernel32):
    kernel32_dict = {}
    kernel32_functions = {}
    kernel32_restypes = {}
    kernel32_argtypes = {}

    # BOOL WINAPI CloseHandle(_In_  HANDLE hObject);
    kernel32_functions[Close_Handle] = kernel32.CloseHandle
    kernel32_restypes[Close_Handle] = BOOL
    kernel32_argtypes[Close_Handle] = [HANDLE]

    # BOOL WINAPI ReadFile(_In_ HANDLE hFile,_Out_ LPVOID lpBuffer,_In_ DWORD nNumberOfBytesToRead,_Out_opt_ LPDWORD lpNumberOfBytesRead,_Inout_opt_ LPOVERLAPPED lpOverlapped);
    kernel32_functions[ReadFile] = kernel32.ReadFile
    kernel32_restypes[ReadFile] = BOOL
    kernel32_argtypes[ReadFile] = [HANDLE, c_void_p, DWORD, POINTER(DWORD), POINTER(Overlapped)]

    # BOOL WINAPI CancelIo(_In_  HANDLE hFile);
    kernel32_functions[CancelIo] = kernel32.CancelIo
    kernel32_restypes[CancelIo] = BOOL
    kernel32_argtypes[CancelIo] = [HANDLE]

    # BOOL WINAPI WriteFile(_In_ HANDLE hFile,_In_ LPCVOID lpBuffer,_In_ DWORD nNumberOfBytesToWrite,_Out_opt_ LPDWORD lpNumberOfBytesWritten,_Inout_opt_  LPOVERLAPPED lpOverlapped);
    kernel32_functions[WriteFile] = kernel32.WriteFile
    kernel32_restypes[WriteFile] = BOOL
    kernel32_argtypes[WriteFile] = [HANDLE, c_void_p, DWORD, POINTER(DWORD), POINTER(Overlapped)]

    # BOOL WINAPI SetEvent(_In_ HANDLE hEvent);
    kernel32_functions[SetEvent] = kernel32.SetEvent
    kernel32_restypes[SetEvent] = BOOL
    kernel32_argtypes[SetEvent] = [HANDLE]

    # DWORD WINAPI WaitForSingleObject(_In_ HANDLE hHandle, _In_  DWORD dwMilliseconds);
    kernel32_functions[WaitForSingleObject] = kernel32.WaitForSingleObject
    kernel32_restypes[WaitForSingleObject] = DWORD
    kernel32_argtypes[WaitForSingleObject] = [HANDLE, DWORD]

    # HANDLE WINAPI CreateFile(_In_ LPCTSTR lpFileName,_In_ DWORD dwDesiredAccess,_In_ DWORD dwShareMode,_In_opt_ LPSECURITY_ATTRIBUTES lpSecurityAttributes,_In_ DWORD dwCreationDisposition,_In_ DWORD dwFlagsAndAttributes,_In_opt_ HANDLE hTemplateFile);
    kernel32_functions[CreateFile] = kernel32.CreateFileW
    kernel32_restypes[CreateFile] = HANDLE

    # DWORD WINAPI GetLastError(void)
    kernel32_functions[GetLastError] = kernel32.GetLastError
    kernel32_restypes[GetLastError] = DWORD
    kernel32_argtypes[GetLastError] = []

    kernel32_dict["functions"] = kernel32_functions
    kernel32_dict["restypes"] = kernel32_restypes
    kernel32_dict["argtypes"] = kernel32_argtypes
    return kernel32_dict


def get_setupapi_functions(setupapi):
    setupapi_dict = {}
    setupapi_functions = {}
    setupapi_restypes = {}
    setupapi_argtypes = {}

    # HDEVINFO SetupDiGetClassDevs(_In_opt_ const GUID *ClassGuid,_In_opt_ PCTSTR Enumerator,_In_opt_ HWND hwndParent,_In_ DWORD Flags);
    setupapi_functions[SetupDiGetClassDevs] = setupapi.SetupDiGetClassDevsW
    setupapi_restypes[SetupDiGetClassDevs] = HANDLE
    setupapi_argtypes[SetupDiGetClassDevs] = [POINTER(GUID), c_wchar_p, HANDLE, DWORD]

    # BOOL SetupDiEnumDeviceInterfaces(_In_ HDEVINFO DeviceInfoSet,_In_opt_ PSP_DEVINFO_DATA DeviceInfoData,_In_ const GUID *InterfaceClassGuid,_In_ DWORD MemberIndex,_Out_ PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData);
    setupapi_functions[SetupDiEnumDeviceInterfaces] = setupapi.SetupDiEnumDeviceInterfaces
    setupapi_restypes[SetupDiEnumDeviceInterfaces] = BOOL
    setupapi_argtypes[SetupDiEnumDeviceInterfaces] = [
        c_void_p,
        POINTER(SpDevinfoData),
        POINTER(GUID),
        DWORD,
        POINTER(SpDeviceInterfaceData),
    ]

    # BOOL SetupDiGetDeviceInterfaceDetail(_In_ HDEVINFO DeviceInfoSet,_In_ PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,_Out_opt_ PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,_In_ DWORD DeviceInterfaceDetailDataSize,_Out_opt_  PDWORD RequiredSize,_Out_opt_  PSP_DEVINFO_DATA DeviceInfoData);
    setupapi_functions[SetupDiGetDeviceInterfaceDetail] = setupapi.SetupDiGetDeviceInterfaceDetailW
    setupapi_restypes[SetupDiGetDeviceInterfaceDetail] = BOOL
    setupapi_argtypes[SetupDiGetDeviceInterfaceDetail] = [
        c_void_p,
        POINTER(SpDeviceInterfaceData),
        POINTER(SpDeviceInterfaceDetailData),
        DWORD,
        POINTER(DWORD),
        POINTER(SpDevinfoData),
    ]

    # BOOL SetupDiGetDeviceInterfaceDetail(_In_ HDEVINFO DeviceInfoSet,_In_ PSP_DEVICE_INTERFACE_DATA DeviceInterfaceData,_Out_opt_ PSP_DEVICE_INTERFACE_DETAIL_DATA DeviceInterfaceDetailData,_In_ DWORD DeviceInterfaceDetailDataSize,_Out_opt_  PDWORD RequiredSize,_Out_opt_  PSP_DEVINFO_DATA DeviceInfoData);
    setupapi_functions[
        SetupDiGetDeviceRegistryProperty
    ] = setupapi.SetupDiGetDeviceRegistryPropertyW
    setupapi_restypes[SetupDiGetDeviceRegistryProperty] = BOOL
    setupapi_argtypes[SetupDiGetDeviceRegistryProperty] = [
        c_void_p,
        POINTER(SpDevinfoData),
        DWORD,
        POINTER(DWORD),
        c_void_p,
        DWORD,
        POINTER(DWORD),
    ]
    # [HDEVINFO, PSP_DEVINFO_DATA, DWORD, PDWORD, PBYTE, DWORD, PDWORD]

    # BOOL SetupDiEnumDeviceInfo(HDEVINFO DeviceInfoSet, DWORD MemberIndex, PSP_DEVINFO_DATA DeviceInfoData);
    setupapi_functions[SetupDiEnumDeviceInfo] = setupapi.SetupDiEnumDeviceInfo
    setupapi_restypes[SetupDiEnumDeviceInfo] = BOOL
    setupapi_argtypes[SetupDiEnumDeviceInfo] = [c_void_p, DWORD, POINTER(SpDevinfoData)]

    setupapi_dict["functions"] = setupapi_functions
    setupapi_dict["restypes"] = setupapi_restypes
    setupapi_dict["argtypes"] = setupapi_argtypes
    return setupapi_dict



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Type

import acconeer.exptool as et

from .calibration import Calibration, CalibrationMapper


class ModuleFamily(Enum):
    EXAMPLE = "Example processing"
    SERVICE = "Services"
    DETECTOR = "Detectors"
    OTHER = None


@dataclass(frozen=True)
class ModuleInfo:
    key: str
    label: str
    pg_updater: Any
    processing_config_class: Any
    module_family: ModuleFamily
    sensor_config_class: Any
    processor: Any
    multi_sensor: Any
    docs_url: Optional[str]
    calibration_class: Optional[Type[Calibration]] = None
    calibration_config_class: Optional[Type[et.configbase.Config]] = None
    calibration_mapper: Optional[Type[CalibrationMapper]] = None
